# gpl: author Silvio Falcinelli. Fixes by angavrilov and others.
# special thanks to user blenderartists.org cmomoney
# -*- coding: utf-8 -*-

import bpy
from os import path as os_path
from bpy.types import Operator
from math import (
        log2, ceil,
        )
from bpy.props import (
        BoolProperty,
        EnumProperty,
        )
from .warning_messages_utils import (
        warning_messages,
        c_is_cycles_addon_enabled,
        c_data_has_materials,
        collect_report,
        )

# -----------------------------------------------------------------------------
# Globals

# switch for operator's function called after AutoNodeInitiate
CHECK_AUTONODE = False

# set the node color for baked images (default greenish)
NODE_COLOR = (0.32, 0.75, 0.32)
# set the node color for the paint base images (default reddish)
NODE_COLOR_PAINT = (0.6, 0.0, 0.0)
# set the mix node color (default blueish)
NODE_COLOR_MIX = (0.1, 0.7, 0.8)

# color for sculpt/texture painting setting (default clay the last entry is Roughness)
PAINT_SC_COLOR = (0.80, 0.75, 0.54, 0.9)
CLAY_GLOSSY = (0.38, 0.032, 0.023, 1)

# -----------------------------------------------------------------------------
# Functions


def AutoNodeSwitch(renderer="CYCLES", switch="OFF", operator=None):
    mats = bpy.data.materials
    use_nodes = (True if switch in ("ON") else False)
    warn_message = ('BI_SW_NODES_ON' if switch in ("ON") else
                    'BI_SW_NODES_OFF')
    warn_message_2 = ('CYC_SW_NODES_ON' if switch in ("ON") else
                      'CYC_SW_NODES_OFF')
    for cmat in mats:
        cmat.use_nodes = use_nodes
    renders = ('CYCLES' if renderer and renderer == "CYCLES" else
               'BLENDER_RENDER')
    bpy.context.scene.render.engine = renders
    if operator:
        warning_messages(operator, (warn_message_2 if renders in ('CYCLES') else
                                    warn_message))


def SetFakeUserTex():
    images = bpy.data.images
    for image in images:
        has_user = getattr(image, "users", -1)
        image_name = getattr(image, "name", "NONAME")

        if has_user == 0:
            image.use_fake_user = True
            collect_report("INFO: Set fake user for unused image: " + image_name)


def BakingText(tex, mode, tex_type=None):
    collect_report("INFO: start bake texture named: " + tex.name)
    saved_img_path = None

    bpy.ops.object.mode_set(mode='OBJECT')
    sc = bpy.context.scene
    tmat = ''
    img = ''
    Robj = bpy.context.active_object
    for n in bpy.data.materials:
        if n.name == 'TMP_BAKING':
            tmat = n
    if not tmat:
        tmat = bpy.data.materials.new('TMP_BAKING')
        tmat.name = "TMP_BAKING"

    bpy.ops.mesh.primitive_plane_add()
    tm = bpy.context.active_object
    tm.name = "TMP_BAKING"
    tm.data.name = "TMP_BAKING"
    bpy.ops.object.select_pattern(extend=False, pattern="TMP_BAKING",
                                  case_sensitive=False)
    sc.objects.active = tm
    bpy.context.scene.render.engine = 'BLENDER_RENDER'
    tm.data.materials.append(tmat)
    if len(tmat.texture_slots.items()) == 0:
        tmat.texture_slots.add()
    tmat.texture_slots[0].texture_coords = 'UV'
    tmat.texture_slots[0].use_map_alpha = True
    tmat.texture_slots[0].texture = tex.texture
    tmat.texture_slots[0].use_map_alpha = True
    tmat.texture_slots[0].use_map_color_diffuse = False
    tmat.use_transparency = True
    tmat.alpha = 0
    tmat.use_nodes = False
    tmat.diffuse_color = 1, 1, 1
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.uv.unwrap()

    # clean up temporary baking images if any
    for n in bpy.data.images:
        if n.name == 'TMP_BAKING':
            n.user_clear()
            bpy.data.images.remove(n)

    if mode == "ALPHA" and tex.texture.type == 'IMAGE':
        sizeX = tex.texture.image.size[0]
        sizeY = tex.texture.image.size[1]
    else:
        bake_size = (int(sc.mat_specials.img_bake_size) if
                     sc.mat_specials.img_bake_size else 1024)
        sizeX = bake_size
        sizeY = bake_size

    bpy.ops.image.new(name="TMP_BAKING", width=sizeX, height=sizeY,
                      color=(0.0, 0.0, 0.0, 1.0), alpha=True, float=False)

    sc.render.engine = 'BLENDER_RENDER'
    img = bpy.data.images["TMP_BAKING"]
    img = bpy.data.images.get("TMP_BAKING")
    img.file_format = ("JPEG" if not mode == "ALPHA" else "PNG")

    # switch temporarly to 'IMAGE EDITOR', other approaches are not reliable
    check_area = False
    store_area = bpy.context.area.type
    collect_report("INFO: Temporarly switching context to Image Editor")
    try:
        bpy.context.area.type = 'IMAGE_EDITOR'
        bpy.context.area.spaces[0].image = bpy.data.images["TMP_BAKING"]
        check_area = True
    except:
        collect_report("ERROR: Setting to Image Editor failed, Baking aborted")
        check_area = False

    if check_area:
        paths = bpy.path.abspath(sc.mat_specials.conv_path)
        tex_name = getattr(getattr(tex.texture, "image", None), "name", None)
        texture_name = (tex_name.rpartition(".")[0] if tex_name else tex.texture.name)
        new_tex_name = "baked"
        name_append = ("_BAKING" if mode == "ALPHA" and
                       tex.texture.type == 'IMAGE' else "_PTEXT")
        new_appendix = (".jpg" if not mode == "ALPHA" else ".png")

        if name_append in texture_name:
            new_tex_name = texture_name
        elif tex_type:
            new_tex_name = tex_type + name_append
        else:
            new_tex_name = texture_name + name_append

        img.filepath_raw = paths + new_tex_name + new_appendix
        saved_img_path = img.filepath_raw

        sc.render.bake_type = 'ALPHA'
        sc.render.use_bake_selected_to_active = True
        sc.render.use_bake_clear = True

        # try to bake if it fails give report
        try:
            bpy.ops.object.bake_image()
            img.save()
        except:
            # no return value, so the image loading is skipped
            saved_img_path = None
            collect_report("ERROR: Baking could not be completed. "
                           "Check System Console for info")
    if store_area:
        bpy.context.area.type = store_area

    bpy.ops.object.mode_set(mode='OBJECT')
    bpy.ops.object.delete()
    bpy.ops.object.select_pattern(extend=False, pattern=Robj.name, case_sensitive=False)
    sc.objects.active = Robj
    img.user_clear()
    bpy.data.images.remove(img)

    if tmat.users == 0:
        bpy.data.materials.remove(tmat)

    if saved_img_path:
        collect_report("------- Baking finished -------")
        return saved_img_path


def AutoNodeInitiate(active=False, operator=None):
    # Checks with bpy.ops.material.check_converter_path
    # if it's possible to write in the output path
    # if it passes procedes with calling AutoNode

    # if CheckImagePath(operator):
    check_path = bpy.ops.material.check_converter_path()

    global CHECK_AUTONODE

    if 'FINISHED' in check_path:
        sc = bpy.context.scene
        CHECK_AUTONODE = True
        collect_report("_______________________", True, False)
        AutoNode(active, operator)
        if sc.mat_specials.SET_FAKE_USER:
            SetFakeUserTex()
    else:
        warning_messages(operator, 'DIR_PATH_CONVERT', override=True)


def AutoNode(active=False, operator=None):
    global CHECK_AUTONODE
    sc = bpy.context.scene
    if active:
        # fix for empty slots by angavrilov
        mats = [slot.material for slot in bpy.context.active_object.material_slots if
                slot.material]
    else:
        mats = bpy.data.materials

    # No Materials for the chosen action - abort
    if not mats:
        CHECK_AUTONODE = False
        if operator:
            if active:
                act_obj = bpy.context.active_object
                warning_messages(operator, 'CONV_NO_OBJ_MAT', act_obj.name)
            else:
                warning_messages(operator, 'CONV_NO_SC_MAT')
        return

    for cmat in mats:
        # check for empty material (it will fall through the first check)
        test_empty = getattr(cmat, "name", None)
        if test_empty is None:
            collect_report("An empty material was hit, skipping")
            continue

        cmat.use_nodes = True
        TreeNodes = cmat.node_tree
        links = TreeNodes.links

        # Don't alter nodes of locked materials
        locked = False
        for n in TreeNodes.nodes:
            if n.type == 'ShaderNodeOutputMaterial':
                if n.label == 'Locked':
                    locked = True
                    break

        if not locked:
            # Convert this material from non-nodes to Cycles nodes
            shader = ''
            shtsl = ''
            Add_Emission = ''
            Add_Translucent = ''
            Mix_Alpha = ''
            sT = False
            # check if some link creation failed
            link_fail = False

            for n in TreeNodes.nodes:
                TreeNodes.nodes.remove(n)

            # Starting point is diffuse BSDF and output material
            # and a Color Ramp node
            shader = TreeNodes.nodes.new('ShaderNodeBsdfDiffuse')
            shader.location = 10, 10
            shader_val = TreeNodes.nodes.new('ShaderNodeValToRGB')
            shader_val.location = 0, -200
            shout = TreeNodes.nodes.new('ShaderNodeOutputMaterial')
            shout.location = 200, 10
            try:
                links.new(shader.outputs[0], shout.inputs[0])
                links.new(shader.inputs[0], shader_val.outputs[0])
            except:
                link_fail = True

            # Create other shader types only sculpt/texture paint mode is False
            sculpt_paint = sc.mat_specials.SCULPT_PAINT
            if sculpt_paint is False:

                cmat_is_transp = cmat.use_transparency and cmat.alpha < 1

                if not cmat.raytrace_mirror.use and not cmat_is_transp:
                    if not shader.type == 'ShaderNodeBsdfDiffuse':
                        collect_report("INFO: Make DIFFUSE shader node for: " + cmat.name)
                        TreeNodes.nodes.remove(shader)
                        shader = TreeNodes.nodes.new('ShaderNodeBsdfDiffuse')
                        shader.location = 10, 10
                        try:
                            links.new(shader.outputs[0], shout.inputs[0])
                        except:
                            link_fail = True

                if cmat.raytrace_mirror.use and cmat.raytrace_mirror.reflect_factor > 0.001 and cmat_is_transp:
                    if not shader.type == 'ShaderNodeBsdfGlass':
                        collect_report("INFO: Make GLASS shader node for: " + cmat.name)
                        TreeNodes.nodes.remove(shader)
                        shader = TreeNodes.nodes.new('ShaderNodeBsdfGlass')
                        shader.location = 0, 100
                        try:
                            links.new(shader.outputs[0], shout.inputs[0])
                        except:
                            link_fail = True

                if cmat.raytrace_mirror.use and not cmat_is_transp and cmat.raytrace_mirror.reflect_factor > 0.001:
                    if not shader.type == 'ShaderNodeBsdfGlossy':
                        collect_report("INFO: Make MIRROR shader node for: " + cmat.name)
                        TreeNodes.nodes.remove(shader)
                        shader = TreeNodes.nodes.new('ShaderNodeBsdfGlossy')
                        shader.location = 0, 10
                        try:
                            links.new(shader.outputs[0], shout.inputs[0])
                        except:
                            link_fail = True

                if cmat.emit > 0.001:
                    if (not shader.type == 'ShaderNodeEmission' and not
                       cmat.raytrace_mirror.reflect_factor > 0.001 and not cmat_is_transp):

                        collect_report("INFO: Mix EMISSION shader node for: " + cmat.name)
                        TreeNodes.nodes.remove(shader)
                        shader = TreeNodes.nodes.new('ShaderNodeEmission')
                        shader.location = 0, 200
                        try:
                            links.new(shader.outputs[0], shout.inputs[0])
                        except:
                            link_fail = True
                    else:
                        if not Add_Emission:
                            collect_report("INFO: Add EMISSION shader node for: " + cmat.name)
                            shout.location = 600, 100
                            Add_Emission = TreeNodes.nodes.new('ShaderNodeAddShader')
                            Add_Emission.location = 370, 100

                            shem = TreeNodes.nodes.new('ShaderNodeEmission')
                            shem.location = 0, 200
                            try:
                                links.new(Add_Emission.outputs[0], shout.inputs[0])
                                links.new(shem.outputs[0], Add_Emission.inputs[1])
                                links.new(shader.outputs[0], Add_Emission.inputs[0])
                            except:
                                link_fail = True

                            shem.inputs['Color'].default_value = (cmat.diffuse_color.r,
                                                                  cmat.diffuse_color.g,
                                                                  cmat.diffuse_color.b, 1)
                            shem.inputs['Strength'].default_value = cmat.emit

                if cmat.translucency > 0.001:
                    collect_report("INFO: Add BSDF_TRANSLUCENT shader node for: " + cmat.name)
                    shout.location = 770, 330
                    Add_Translucent = TreeNodes.nodes.new('ShaderNodeAddShader')
                    Add_Translucent.location = 580, 490

                    shtsl = TreeNodes.nodes.new('ShaderNodeBsdfTranslucent')
                    shtsl.location = 400, 350
                    try:
                        links.new(Add_Translucent.outputs[0], shout.inputs[0])
                        links.new(shtsl.outputs[0], Add_Translucent.inputs[1])

                        if Add_Emission:
                            links.new(Add_Emission.outputs[0], Add_Translucent.inputs[0])
                        else:
                            links.new(shader.outputs[0], Add_Translucent.inputs[0])
                    except:
                        link_fail = True

                    shtsl.inputs['Color'].default_value = (cmat.translucency,
                                                           cmat.translucency,
                                                           cmat.translucency, 1)
            if sculpt_paint is False:
                shader.inputs['Color'].default_value = (cmat.diffuse_color.r,
                                                        cmat.diffuse_color.g,
                                                        cmat.diffuse_color.b, 1)
            else:
                # Create Clay Material (Diffuse, Glossy, Layer Weight)
                shader.inputs['Color'].default_value = PAINT_SC_COLOR
                shader.inputs['Roughness'].default_value = 0.9

                # remove Color Ramp and links from the default shader and reroute
                try:
                    shout.location = 400, 0
                    for link in links:
                        links.remove(link)

                    clay_frame = TreeNodes.nodes.new('NodeFrame')
                    clay_frame.name = 'Clay Material'
                    clay_frame.label = 'Clay Material'

                    sh_glossy = TreeNodes.nodes.new('ShaderNodeBsdfGlossy')
                    sh_glossy.location = 0, 200
                    sh_glossy.inputs['Color'].default_value = CLAY_GLOSSY
                    sh_mix = TreeNodes.nodes.new('ShaderNodeMixShader')
                    sh_mix.location = 200, 0
                    sh_weight = TreeNodes.nodes.new('ShaderNodeLayerWeight')
                    sh_weight.location = 0, 350
                    links.new(sh_mix.outputs[0], shout.inputs[0])
                    links.new(sh_weight.outputs[1], sh_mix.inputs[0])
                    links.new(shader.outputs[0], sh_mix.inputs[1])
                    links.new(sh_glossy.outputs[0], sh_mix.inputs[2])
                    # set frame as parent to everything
                    for clay_node in (shader, sh_glossy, sh_mix, sh_weight):
                        clay_node.parent = clay_frame
                except:
                    collect_report("ERROR: Failure to create Clay Material")

            if not sculpt_paint:
                if shader.type == 'ShaderNodeBsdfDiffuse':
                    shader.inputs['Roughness'].default_value = cmat.specular_intensity

                if shader.type == 'ShaderNodeBsdfGlossy':
                    shader.inputs['Roughness'].default_value = 1 - cmat.raytrace_mirror.gloss_factor

                if shader.type == 'ShaderNodeBsdfGlass':
                    shader.inputs['Roughness'].default_value = 1 - cmat.raytrace_mirror.gloss_factor
                    shader.inputs['IOR'].default_value = cmat.raytrace_transparency.ior

                if shader.type == 'ShaderNodeEmission':
                    shader.inputs['Strength'].default_value = cmat.emit

            # texture presence check
            is_textures = False

            for tex in cmat.texture_slots:
                if tex:
                    if not is_textures:
                        is_textures = True
                        break

            if is_textures:
                # collect the texture nodes created
                # for spreading a bit the texture nodes
                tex_node_collect = []

                sM = True

                for tex in cmat.texture_slots:
                    sT = False
                    tex_use = getattr(tex, "use", None)
                    baked_path = None
                    if tex_use:
                        tex_node_loc = -200, 450
                        ma_alpha = getattr(tex, "use_map_alpha", None)
                        sM = (False if ma_alpha else True)
                        img = None

                        if tex.texture.type == 'IMAGE':
                            if sc.mat_specials.EXTRACT_ALPHA and tex.texture.use_alpha:
                                if (not
                                   os_path.exists(bpy.path.abspath(tex.texture.image.filepath + "_BAKING.png")) or
                                   sc.mat_specials.EXTRACT_OW):
                                    baked_path = BakingText(tex, 'ALPHA')

                                    if baked_path:
                                        try:
                                            img = bpy.data.images.load(baked_path)
                                            collect_report("INFO: Loading Baked texture path:")
                                            collect_report(baked_path)
                                        except:
                                            collect_report("ERROR: Baked image could not be loaded")
                            else:
                                has_image = getattr(tex.texture, "image", None)
                                if has_image:
                                    img = has_image

                            if img:
                                img_name = (img.name if hasattr(img, "name") else "NO NAME")
                                shtext = TreeNodes.nodes.new('ShaderNodeTexImage')
                                shtext.location = tex_node_loc
                                shtext.hide = True
                                shtext.width_hidden = 150
                                shtext.image = img
                                shtext.name = img_name
                                shtext.label = "Image " + img_name
                                if baked_path:
                                    shtext.use_custom_color = True
                                    shtext.color = NODE_COLOR
                                collect_report("INFO: Creating Image Node for image: " + img_name)
                                tex_node_collect.append(shtext)
                                sT = True
                            else:
                                collect_report("ERROR: A problem occured with loading an image for {} "
                                               "(possibly missing)".format(tex.texture.name))
                        else:
                            if sc.mat_specials.EXTRACT_PTEX or (sc.mat_specials.EXTRACT_ALPHA and ma_alpha):
                                if (not os_path.exists(bpy.path.abspath(tex.texture.name + "_PTEXT.jpg")) or
                                   sc.mat_specials.EXTRACT_OW):
                                    tex_type = tex.texture.type.lower()
                                    collect_report("Attempting to Extract Procedural Texture type: " + tex_type)
                                    baked_path = BakingText(tex, 'PTEX', tex_type)

                                if baked_path:
                                    try:
                                        img = bpy.data.images.load(baked_path)
                                        collect_report("Loading Baked texture path:")
                                        collect_report(baked_path)
                                        img_name = (img.name if hasattr(img, "name") else "NO NAME")
                                        shtext = TreeNodes.nodes.new('ShaderNodeTexImage')
                                        shtext.location = tex_node_loc
                                        shtext.hide = True
                                        shtext.width_hidden = 150
                                        shtext.image = img
                                        shtext.name = img_name
                                        shtext.label = "Baked Image " + img_name
                                        shtext.use_custom_color = True
                                        shtext.color = NODE_COLOR
                                        collect_report("Creating Image Node for baked image: " + img_name)
                                        tex_node_collect.append(shtext)
                                        sT = True
                                    except:
                                        collect_report("ERROR: Failure to load baked image: " + img_name)
                                else:
                                    collect_report("ERROR: Failure during baking, no images loaded")

                    if sculpt_paint is False:
                        if (cmat_is_transp and cmat.raytrace_transparency.ior == 1 and
                           not cmat.raytrace_mirror.use and sM):

                            if not shader.type == 'ShaderNodeBsdfTransparent':
                                collect_report("INFO: Make TRANSPARENT shader node for: " + cmat.name)
                                TreeNodes.nodes.remove(shader)
                                shader = TreeNodes.nodes.new('ShaderNodeBsdfTransparent')
                                shader.location = 0, 470
                                try:
                                    links.new(shader.outputs[0], shout.inputs[0])
                                except:
                                    link_fail = True

                                shader.inputs['Color'].default_value = (cmat.diffuse_color.r,
                                                                        cmat.diffuse_color.g,
                                                                        cmat.diffuse_color.b, 1)

                    if sT and sculpt_paint is False:
                        if tex.use_map_color_diffuse:
                            try:
                                links.new(shtext.outputs[0], shader.inputs[0])
                            except:
                                pass
                        if tex.use_map_emit:
                            if not Add_Emission:
                                collect_report("INFO: Mix EMISSION + Texture shader node for: " + cmat.name)
                                intensity = 0.5 + (tex.emit_factor / 2)

                                shout.location = 550, 330
                                Add_Emission = TreeNodes.nodes.new('ShaderNodeAddShader')
                                Add_Emission.name = "Add_Emission"
                                Add_Emission.location = 370, 490

                                shem = TreeNodes.nodes.new('ShaderNodeEmission')
                                shem.location = 180, 380

                                try:
                                    links.new(Add_Emission.outputs[0], shout.inputs[0])
                                    links.new(shem.outputs[0], Add_Emission.inputs[1])
                                    links.new(shader.outputs[0], Add_Emission.inputs[0])
                                    links.new(shtext.outputs[0], shem.inputs[0])
                                except:
                                    link_fail = True

                                shem.inputs['Color'].default_value = (cmat.diffuse_color.r,
                                                                      cmat.diffuse_color.g,
                                                                      cmat.diffuse_color.b, 1)
                                shem.inputs['Strength'].default_value = intensity * 2

                        if tex.use_map_mirror:
                            try:
                                links.new(shtext.outputs[0], shader.inputs[0])
                            except:
                                link_fail = True

                        if tex.use_map_translucency:
                            if not Add_Translucent:
                                collect_report("INFO: Add Translucency + Texture shader node for: " + cmat.name)

                                intensity = 0.5 + (tex.emit_factor / 2)
                                shout.location = 550, 330
                                Add_Translucent = TreeNodes.nodes.new('ShaderNodeAddShader')
                                Add_Translucent.name = "Add_Translucent"
                                Add_Translucent.location = 370, 290

                                shtsl = TreeNodes.nodes.new('ShaderNodeBsdfTranslucent')
                                shtsl.location = 180, 240
                                try:
                                    links.new(shtsl.outputs[0], Add_Translucent.inputs[1])

                                    if Add_Emission:
                                        links.new(Add_Translucent.outputs[0], shout.inputs[0])
                                        links.new(Add_Emission.outputs[0], Add_Translucent.inputs[0])
                                        pass
                                    else:
                                        links.new(Add_Translucent.outputs[0], shout.inputs[0])
                                        links.new(shader.outputs[0], Add_Translucent.inputs[0])

                                    links.new(shtext.outputs[0], shtsl.inputs[0])
                                except:
                                    link_fail = True

                        if tex.use_map_alpha:
                            if not Mix_Alpha:
                                collect_report("INFO: Mix Alpha + Texture shader node for: " + cmat.name)

                                shout.location = 750, 330
                                Mix_Alpha = TreeNodes.nodes.new('ShaderNodeMixShader')
                                Mix_Alpha.name = "Add_Alpha"
                                Mix_Alpha.location = 570, 290
                                sMask = TreeNodes.nodes.new('ShaderNodeBsdfTransparent')
                                sMask.location = 250, 180
                                tMask, imask = None, None

                                # search if the texture node already exists, if not create
                                nodes = getattr(cmat.node_tree, "nodes", None)
                                img_name = getattr(img, "name", "NO NAME")
                                for node in nodes:
                                    if type(node) == bpy.types.ShaderNodeTexImage:
                                        node_name = getattr(node, "name")
                                        if img_name in node_name:
                                            tMask = node
                                            collect_report("INFO: Using existing Texture Node for Mask: " + node_name)
                                            break

                                if tMask is None:
                                    tMask = TreeNodes.nodes.new('ShaderNodeTexImage')
                                    tMask.location = tex_node_loc
                                    tex_node_collect.append(tMask)

                                    try:
                                        file_path = getattr(img, "filepath", None)
                                        if file_path:
                                            imask = bpy.data.images.load(file_path)
                                        else:
                                            imask = bpy.data.images.get(img_name)
                                        collect_report("INFO: Attempting to load image for Mask: " + img_name)
                                    except:
                                        collect_report("ERROR: Failure to load image for Mask: " + img_name)

                                    if imask:
                                        tMask.image = imask

                                if tMask:
                                    try:
                                        links.new(Mix_Alpha.inputs[0], tMask.outputs[1])
                                        links.new(shout.inputs[0], Mix_Alpha.outputs[0])
                                        links.new(sMask.outputs[0], Mix_Alpha.inputs[1])

                                        if not Add_Translucent:
                                            if Add_Emission:
                                                links.new(Mix_Alpha.inputs[2], Add_Emission.outputs[0])
                                            else:
                                                links.new(Mix_Alpha.inputs[2], shader.outputs[0])
                                        else:
                                            links.new(Mix_Alpha.inputs[2], Add_Translucent.outputs[0])
                                    except:
                                        link_fail = True
                                else:
                                    collect_report("ERROR: Mix Alpha could not be created "
                                                   "(mask image could not be loaded)")

                        if tex.use_map_normal:
                            t = TreeNodes.nodes.new('ShaderNodeRGBToBW')
                            t.location = -0, 300
                            try:
                                links.new(t.outputs[0], shout.inputs[2])
                                links.new(shtext.outputs[1], t.inputs[0])
                            except:
                                link_fail = True

                if sculpt_paint:
                    try:
                        # create a new image for texture painting and make it active
                        img_size = (int(sc.mat_specials.img_bake_size) if
                                    sc.mat_specials.img_bake_size else 1024)
                        paint_mat_name = getattr(cmat, "name", "NO NAME")
                        paint_img_name = "Paint Base Image {}".format(paint_mat_name)
                        bpy.ops.image.new(name=paint_img_name, width=img_size, height=img_size,
                                          color=(1.0, 1.0, 1.0, 1.0), alpha=True, float=False)

                        img = bpy.data.images.get(paint_img_name)
                        img_name = (img.name if hasattr(img, "name") else "NO NAME")
                        shtext = TreeNodes.nodes.new('ShaderNodeTexImage')
                        shtext.hide = True
                        shtext.width_hidden = 150
                        shtext.location = tex_node_loc
                        shtext.image = img
                        shtext.name = img_name
                        shtext.label = "Paint: " + img_name
                        shtext.use_custom_color = True
                        shtext.color = NODE_COLOR_PAINT
                        shtext.select = True
                        collect_report("INFO: Creating Image Node for Painting: " + img_name)
                        collect_report("WARNING: Don't forget to save it on Disk")
                        tex_node_collect.append(shtext)
                    except:
                        collect_report("ERROR: Failed to create image and node for Texture Painting")

                # spread the texture nodes, create node frames if necessary
                # create texture coordinate and mapping too
                row_node = -1
                tex_node_collect_size = len(tex_node_collect)
                median_point = ((tex_node_collect_size / 2) * 100)
                check_frame = bool(tex_node_collect_size > 1)

                node_frame, tex_map = None, None
                node_f_coord, node_f_mix = None, None
                tex_map_collection, tex_map_coord = [], None
                tree_size, tree_tex_start = 0, 0

                if check_frame:
                    node_frame = TreeNodes.nodes.new('NodeFrame')
                    node_frame.name = 'Converter Textures'
                    node_frame.label = 'Converter Textures'

                    node_f_coord = TreeNodes.nodes.new('NodeFrame')
                    node_f_coord.name = "Coordinates"
                    node_f_coord.label = "Coordinates"

                    node_f_mix = TreeNodes.nodes.new('NodeFrame')
                    node_f_mix.name = "Mix"
                    node_f_mix.label = "Mix"

                if tex_node_collect:
                    tex_map_coord = TreeNodes.nodes.new('ShaderNodeTexCoord')
                    tex_map_coord.location = -900, 575

                    # precalculate the depth of the inverted tree
                    tree_size = int(ceil(log2(tex_node_collect_size)))
                    # offset the start of the mix nodes by the depth of the tree
                    tree_tex_start = ((tree_size + 1) * 150)

                for node_tex in tex_node_collect:
                    row_node += 1
                    col_node_start = (median_point - (-(row_node * 50) + median_point))
                    tex_node_row = tree_tex_start + 300
                    mix_node_row = tree_tex_start + 620
                    tex_node_loc = (-(tex_node_row), col_node_start)

                    try:
                        node_tex.location = tex_node_loc
                        if check_frame:
                            node_tex.parent = node_frame
                        else:
                            node_tex.hide = False

                        tex_node_name = getattr(node_tex, "name", "NO NAME")
                        tex_map_name = "Mapping: {}".format(tex_node_name)
                        tex_map = TreeNodes.nodes.new('ShaderNodeMapping')
                        tex_map.location = (-(mix_node_row), col_node_start)
                        tex_map.width = 240
                        tex_map.hide = True
                        tex_map.width_hidden = 150
                        tex_map.name = tex_map_name
                        tex_map.label = tex_map_name
                        tex_map_collection.append(tex_map)
                        links.new(tex_map.outputs[0], node_tex.inputs[0])
                    except:
                        link_fail = True
                        continue

                if tex_map_collection:
                    tex_mix_start = len(tex_map_collection) / 2
                    row_map_start = -(tree_tex_start + 850)

                    if tex_map_coord:
                        tex_map_coord.location = (row_map_start,
                                                  (median_point - (tex_mix_start * 50)))

                    for maps in tex_map_collection:
                        try:
                            if node_f_coord:
                                maps.parent = node_f_coord
                            else:
                                maps.hide = False

                            links.new(maps.inputs[0], tex_map_coord.outputs['UV'])
                        except:
                            link_fail = True
                            continue

                # create mix nodes to connect texture nodes to the shader input
                # sculpt mode doesn't need them
                if check_frame and not sculpt_paint:
                    mix_node_pairs = loop_node_from_list(TreeNodes, links, tex_node_collect,
                                                         0, tree_tex_start, median_point, node_f_mix)

                    for n in range(1, tree_size):
                        mix_node_pairs = loop_node_from_list(TreeNodes, links, mix_node_pairs,
                                                             n, tree_tex_start, median_point, node_f_mix)
                    try:
                        for node in mix_node_pairs:
                            links.new(node.outputs[0], shader.inputs[0])
                    except:
                        link_fail = True

                    mix_node_pairs = []

                tex_node_collect, tex_map_collection = [], []

                if link_fail:
                    collect_report("ERROR: Some of the node links failed to connect")

            else:
                collect_report("No textures in the Scene, no Image Nodes to add")

    bpy.context.scene.render.engine = 'CYCLES'


def loop_node_from_list(TreeNodes, links, node_list, loc, start, median_point, frame):
    row = 1
    mix_nodes = []
    node_list_size = len(node_list)
    tuplify = [tuple(node_list[s:s + 2]) for s in range(0, node_list_size, 2)]
    for nodes in tuplify:
        row += 1
        create_mix = create_mix_node(TreeNodes, links, nodes, loc, start,
                                     median_point, row, frame)
        if create_mix:
            mix_nodes.append(create_mix)
    return mix_nodes


def create_mix_node(TreeNodes, links, nodes, loc, start, median_point, row, frame):
    mix_node = TreeNodes.nodes.new('ShaderNodeMixRGB')
    mix_node.name = "MIX level: " + str(loc)
    mix_node.label = "MIX level: " + str(loc)
    mix_node.use_custom_color = True
    mix_node.color = NODE_COLOR_MIX
    mix_node.hide = True
    mix_node.width_hidden = 75

    if frame:
        mix_node.parent = frame
    mix_node.location = -(start - loc * 175), ((median_point / 4) + (row * 50))

    try:
        if len(nodes) > 1:
            links.new(nodes[0].outputs[0], mix_node.inputs["Color2"])
            links.new(nodes[1].outputs[0], mix_node.inputs["Color1"])
        elif len(nodes) == 1:
            links.new(nodes[0].outputs[0], mix_node.inputs["Color1"])
    except:
        collect_report("ERROR: Link failed for mix node {}".format(mix_node.label))
    return mix_node


# -----------------------------------------------------------------------------
# Operator Classes

class mllock(Operator):
    bl_idname = "ml.lock"
    bl_label = "Lock"
    bl_description = "Lock/unlock this material against modification by conversions"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return (c_is_cycles_addon_enabled() and c_data_has_materials())

    def execute(self, context):
        cmat = bpy.context.selected_objects[0].active_material
        TreeNodes = cmat.node_tree
        for n in TreeNodes.nodes:
            if n.type == 'ShaderNodeOutputMaterial':
                if n.label == 'Locked':
                    n.label = ''
                else:
                    n.label = 'Locked'
        return {'FINISHED'}


class mlrefresh(Operator):
    bl_idname = "ml.refresh"
    bl_label = "Convert All Materials"
    bl_description = ("Convert All Materials in the scene from non-nodes to Cycles\n"
                      "Needs saving the .blend file first")
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return (bpy.data.filepath != ""and c_is_cycles_addon_enabled() and
                c_data_has_materials())

    def execute(self, context):
        AutoNodeInitiate(False, self)

        if CHECK_AUTONODE is True:
            enable_unwrap = bpy.context.scene.mat_specials.UV_UNWRAP
            if enable_unwrap:
                obj_name = getattr(context.active_object, "name", "UNNAMED OBJECT")
                try:
                    # it's possible to the active object would fail UV Unwrap
                    bpy.ops.object.editmode_toggle()
                    bpy.ops.uv.unwrap(method='ANGLE_BASED', margin=0.001)
                    bpy.ops.object.editmode_toggle()
                    collect_report("INFO: UV Unwrapping active object "
                                   "{}".format(obj_name))
                except:
                    collect_report("ERROR: UV Unwrapping failed for "
                                   "active object {}".format(obj_name))

        collect_report("Conversion finished !", False, True)
        return {'FINISHED'}


class mlrefresh_active(Operator):
    bl_idname = "ml.refresh_active"
    bl_label = "Convert All Materials From Active Object"
    bl_description = ("Convert all Active Object's Materials from non-nodes to Cycles\n"
                      "Needs saving the .blend file first")
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return (bpy.data.filepath != "" and c_is_cycles_addon_enabled() and
                c_data_has_materials() and context.active_object is not None)

    def execute(self, context):
        AutoNodeInitiate(True, self)
        if CHECK_AUTONODE is True:
            obj_name = getattr(context.active_object, "name", "UNNAMED OBJECT")
            enable_unwrap = bpy.context.scene.mat_specials.UV_UNWRAP
            if enable_unwrap:
                try:
                    # you can already guess it, what could happen here
                    bpy.ops.object.editmode_toggle()
                    bpy.ops.uv.unwrap(method='ANGLE_BASED', margin=0.001)
                    bpy.ops.object.editmode_toggle()
                    collect_report("INFO: UV Unwrapping object {}".format(obj_name))
                except:
                    collect_report("ERROR: UV Unwrapping failed for "
                                   "object {}".format(obj_name))

        collect_report("Conversion finished !", False, True)
        return {'FINISHED'}


class mlrestore(Operator):
    bl_idname = "ml.restore"
    bl_label = "Switch Between Renderers"
    bl_description = ("Switch between Renderers \n"
                      "(Doesn't create new nor converts existing materials)")
    bl_options = {'REGISTER', 'UNDO'}

    switcher = BoolProperty(
            name="Use Nodes",
            description="When restoring, switch Use Nodes On/Off",
            default=True
            )
    renderer = EnumProperty(
            name="Renderer",
            description="Choose Cycles or Blender Internal",
            items=(('CYCLES', "Cycles", "Switch to Cycles"),
                   ('BI', "Blender Internal", "Switch to Blender Internal")),
            default='CYCLES',
            )

    @classmethod
    def poll(cls, context):
        return c_is_cycles_addon_enabled()

    def execute(self, context):
        if self.switcher:
            AutoNodeSwitch(self.renderer, "ON", self)
        else:
            AutoNodeSwitch(self.renderer, "OFF", self)
        return {'FINISHED'}


def register():
    bpy.utils.register_module(__name__)
    pass


def unregister():
    bpy.utils.unregister_module(__name__)
    pass


if __name__ == "__main__":
    register()



#====================== BEGIN GPL LICENSE BLOCK ======================
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#======================= END GPL LICENSE BLOCK ========================

# <pep8 compliant>

bl_info = {
    "name": "Rigify",
    "version": (0, 5),
    "author": "Nathan Vegdahl, Lucio Rossi, Ivan Cappiello",
    "blender": (2, 78, 0),
    "description": "Automatic rigging from building-block components",
    "location": "Armature properties, Bone properties, View3d tools panel, Armature Add menu",
    "wiki_url": "http://wiki.blender.org/index.php/Extensions:2.5/Py/"
                "Scripts/Rigging/Rigify",
    "category": "Rigging"}


if "bpy" in locals():
    import importlib
    importlib.reload(generate)
    importlib.reload(ui)
    importlib.reload(utils)
    importlib.reload(metarig_menu)
    importlib.reload(rig_lists)
else:
    from . import utils, rig_lists, generate, ui, metarig_menu

import bpy
import sys
import os
from bpy.types import AddonPreferences
from bpy.props import BoolProperty


class RigifyPreferences(AddonPreferences):
    # this must match the addon name, use '__package__'
    # when defining this in a submodule of a python package.
    bl_idname = __name__

    def update_legacy(self, context):
        if self.legacy_mode:

            if 'ui' in globals() and 'legacy' in str(globals()['ui']):    # already in legacy mode. needed when rigify is reloaded
                return
            else:
                rigify_dir = os.path.dirname(os.path.realpath(__file__))
                if rigify_dir not in sys.path:
                    sys.path.append(rigify_dir)

                unregister()

                globals().pop('utils')
                globals().pop('rig_lists')
                globals().pop('generate')
                globals().pop('ui')
                globals().pop('metarig_menu')

                import legacy.utils
                import legacy.rig_lists
                import legacy.generate
                import legacy.ui
                import legacy.metarig_menu

                print("ENTERING RIGIFY LEGACY\r\n")

                globals()['utils'] = legacy.utils
                globals()['rig_lists'] = legacy.rig_lists
                globals()['generate'] = legacy.generate
                globals()['ui'] = legacy.ui
                globals()['metarig_menu'] = legacy.metarig_menu

                register()

        else:

            rigify_dir = os.path.dirname(os.path.realpath(__file__))

            if rigify_dir in sys.path:
                id = sys.path.index(rigify_dir)
                sys.path.pop(id)

            unregister()

            globals().pop('utils')
            globals().pop('rig_lists')
            globals().pop('generate')
            globals().pop('ui')
            globals().pop('metarig_menu')

            from . import utils
            from . import rig_lists
            from . import generate
            from . import ui
            from . import metarig_menu

            print("EXIT RIGIFY LEGACY\r\n")

            globals()['utils'] = utils
            globals()['rig_lists'] = rig_lists
            globals()['generate'] = generate
            globals()['ui'] = ui
            globals()['metarig_menu'] = metarig_menu

            register()

    legacy_mode = BoolProperty(
        name='Rigify Legacy Mode',
        description='Select if you want to use Rigify in legacy mode',
        default=False,
        update=update_legacy
    )

    show_expanded = BoolProperty()

    def draw(self, context):
        layout = self.layout
        column = layout.column()
        box = column.box()

        # first stage
        expand = getattr(self, 'show_expanded')
        icon = 'TRIA_DOWN' if expand else 'TRIA_RIGHT'
        col = box.column()
        row = col.row()
        sub = row.row()
        sub.context_pointer_set('addon_prefs', self)
        sub.alignment = 'LEFT'
        op = sub.operator('wm.context_toggle', text='', icon=icon,
                          emboss=False)
        op.data_path = 'addon_prefs.show_expanded'
        sub.label('{}: {}'.format('Rigify', 'Enable Legacy Mode'))
        sub = row.row()
        sub.alignment = 'RIGHT'
        sub.prop(self, 'legacy_mode')

        if expand:
            split = col.row().split(percentage=0.15)
            split.label('Description:')
            split.label(text='When enabled the add-on will run in legacy mode using the old 2.76b feature set.')

        row = layout.row()
        row.label("End of Rigify Preferences")


class RigifyName(bpy.types.PropertyGroup):
    name = bpy.props.StringProperty()


class RigifyColorSet(bpy.types.PropertyGroup):
    name = bpy.props.StringProperty(name="Color Set", default=" ")
    active = bpy.props.FloatVectorProperty(
                                   name="object_color",
                                   subtype='COLOR',
                                   default=(1.0, 1.0, 1.0),
                                   min=0.0, max=1.0,
                                   description="color picker"
                                   )
    normal = bpy.props.FloatVectorProperty(
                                   name="object_color",
                                   subtype='COLOR',
                                   default=(1.0, 1.0, 1.0),
                                   min=0.0, max=1.0,
                                   description="color picker"
                                   )
    select = bpy.props.FloatVectorProperty(
                                   name="object_color",
                                   subtype='COLOR',
                                   default=(1.0, 1.0, 1.0),
                                   min=0.0, max=1.0,
                                   description="color picker"
                                   )
    standard_colors_lock = bpy.props.BoolProperty(default=True)


class RigifySelectionColors(bpy.types.PropertyGroup):

    select = bpy.props.FloatVectorProperty(
                                           name="object_color",
                                           subtype='COLOR',
                                           default=(0.314, 0.784, 1.0),
                                           min=0.0, max=1.0,
                                           description="color picker"
                                           )

    active = bpy.props.FloatVectorProperty(
                                           name="object_color",
                                           subtype='COLOR',
                                           default=(0.549, 1.0, 1.0),
                                           min=0.0, max=1.0,
                                           description="color picker"
                                           )


class RigifyParameters(bpy.types.PropertyGroup):
    name = bpy.props.StringProperty()


class RigifyArmatureLayer(bpy.types.PropertyGroup):

    def get_group(self):
        if 'group_prop' in self.keys():
            return self['group_prop']
        else:
            return 0

    def set_group(self, value):
        arm = bpy.context.object.data
        if value > len(arm.rigify_colors):
            self['group_prop'] = len(arm.rigify_colors)
        else:
            self['group_prop'] = value

    name = bpy.props.StringProperty(name="Layer Name", default=" ")
    row = bpy.props.IntProperty(name="Layer Row", default=1, min=1, max=32, description='UI row for this layer')
    set = bpy.props.BoolProperty(name="Selection Set", default=False, description='Add Selection Set for this layer')
    group = bpy.props.IntProperty(name="Bone Group", default=0, min=0, max=32,
                                  get=get_group, set=set_group, description='Assign Bone Group to this layer')

##### REGISTER #####

def register():
    ui.register()
    metarig_menu.register()

    bpy.utils.register_class(RigifyName)
    bpy.utils.register_class(RigifyParameters)

    bpy.utils.register_class(RigifyColorSet)
    bpy.utils.register_class(RigifySelectionColors)
    bpy.utils.register_class(RigifyArmatureLayer)
    bpy.utils.register_class(RigifyPreferences)
    bpy.types.Armature.rigify_layers = bpy.props.CollectionProperty(type=RigifyArmatureLayer)

    bpy.types.PoseBone.rigify_type = bpy.props.StringProperty(name="Rigify Type", description="Rig type for this bone")
    bpy.types.PoseBone.rigify_parameters = bpy.props.PointerProperty(type=RigifyParameters)

    bpy.types.Armature.rigify_colors = bpy.props.CollectionProperty(type=RigifyColorSet)

    bpy.types.Armature.rigify_selection_colors = bpy.props.PointerProperty(type=RigifySelectionColors)

    bpy.types.Armature.rigify_colors_index = bpy.props.IntProperty(default=-1)
    bpy.types.Armature.rigify_colors_lock = bpy.props.BoolProperty(default=True)
    bpy.types.Armature.rigify_theme_to_add = bpy.props.EnumProperty(items=(('THEME01', 'THEME01', ''),
                                                                          ('THEME02', 'THEME02', ''),
                                                                          ('THEME03', 'THEME03', ''),
                                                                          ('THEME04', 'THEME04', ''),
                                                                          ('THEME05', 'THEME05', ''),
                                                                          ('THEME06', 'THEME06', ''),
                                                                          ('THEME07', 'THEME07', ''),
                                                                          ('THEME08', 'THEME08', ''),
                                                                          ('THEME09', 'THEME09', ''),
                                                                          ('THEME10', 'THEME10', ''),
                                                                          ('THEME11', 'THEME11', ''),
                                                                          ('THEME12', 'THEME12', ''),
                                                                          ('THEME13', 'THEME13', ''),
                                                                          ('THEME14', 'THEME14', ''),
                                                                          ('THEME15', 'THEME15', ''),
                                                                          ('THEME16', 'THEME16', ''),
                                                                          ('THEME17', 'THEME17', ''),
                                                                          ('THEME18', 'THEME18', ''),
                                                                          ('THEME19', 'THEME19', ''),
                                                                          ('THEME20', 'THEME20', '')
                                                                           ), name='Theme')

    IDStore = bpy.types.WindowManager
    IDStore.rigify_collection = bpy.props.EnumProperty(items=rig_lists.col_enum_list, default="All",
                                                       name="Rigify Active Collection",
                                                       description="The selected rig collection")

    IDStore.rigify_types = bpy.props.CollectionProperty(type=RigifyName)
    IDStore.rigify_active_type = bpy.props.IntProperty(name="Rigify Active Type", description="The selected rig type")

    IDStore.rigify_advanced_generation = bpy.props.BoolProperty(name="Advanced Options",
                                                                description="Enables/disables advanced options for Rigify rig generation",
                                                                default=False)

    def update_mode(self, context):
        if self.rigify_generate_mode == 'new':
            self.rigify_force_widget_update = False

    IDStore.rigify_generate_mode = bpy.props.EnumProperty(name="Rigify Generate Rig Mode",
                                                          description="'Generate Rig' mode. In 'overwrite' mode the features of the target rig will be updated as defined by the metarig. In 'new' mode a new rig will be created as defined by the metarig. Current mode",
                                                          update=update_mode,
                                                          items=(('overwrite', 'overwrite', ''),
                                                                 ('new', 'new', '')))

    IDStore.rigify_force_widget_update = bpy.props.BoolProperty(name="Force Widget Update",
                                                                description="Forces Rigify to delete and rebuild all the rig widgets. if unset, only missing widgets will be created",
                                                                default=False)

    IDStore.rigify_target_rigs = bpy.props.CollectionProperty(type=RigifyName)
    IDStore.rigify_target_rig = bpy.props.StringProperty(name="Rigify Target Rig",
                                                         description="Defines which rig to overwrite. If unset, a new one called 'rig' will be created",
                                                         default="")

    IDStore.rigify_rig_uis = bpy.props.CollectionProperty(type=RigifyName)
    IDStore.rigify_rig_ui = bpy.props.StringProperty(name="Rigify Target Rig UI",
                                                         description="Defines the UI to overwrite. It should always be the same as the target rig. If unset, 'rig_ui.py' will be used",
                                                         default="")

    IDStore.rigify_rig_basename = bpy.props.StringProperty(name="Rigify Rig Name",
                                                     description="Defines the name of the Rig. If unset, in 'new' mode 'rig' will be used, in 'overwrite' mode the target rig name will be used",
                                                     default="")

    IDStore.rigify_transfer_only_selected = bpy.props.BoolProperty(name="Transfer Only Selected", description="Transfer selected bones only", default=True)
    IDStore.rigify_transfer_start_frame = bpy.props.IntProperty(name="Start Frame", description="First Frame to Transfer", default=0, min= 0)
    IDStore.rigify_transfer_end_frame = bpy.props.IntProperty(name="End Frame", description="Last Frame to Transfer", default=0, min= 0)

    if (ui and 'legacy' in str(ui)) or bpy.context.user_preferences.addons['rigify'].preferences.legacy_mode:
        # update legacy on restart or reload
        bpy.context.user_preferences.addons['rigify'].preferences.legacy_mode = True

    # Add rig parameters
    for rig in rig_lists.rig_list:
        r = utils.get_rig_type(rig)
        try:
            r.add_parameters(RigifyParameters)
        except AttributeError:
            pass


def unregister():
    del bpy.types.PoseBone.rigify_type
    del bpy.types.PoseBone.rigify_parameters

    IDStore = bpy.types.WindowManager
    del IDStore.rigify_collection
    del IDStore.rigify_types
    del IDStore.rigify_active_type
    del IDStore.rigify_advanced_generation
    del IDStore.rigify_generate_mode
    del IDStore.rigify_force_widget_update
    del IDStore.rigify_target_rig
    del IDStore.rigify_target_rigs
    del IDStore.rigify_rig_uis
    del IDStore.rigify_rig_ui
    del IDStore.rigify_rig_basename
    del IDStore.rigify_transfer_only_selected
    del IDStore.rigify_transfer_start_frame
    del IDStore.rigify_transfer_end_frame

    bpy.utils.unregister_class(RigifyName)
    bpy.utils.unregister_class(RigifyParameters)

    bpy.utils.unregister_class(RigifyColorSet)
    bpy.utils.unregister_class(RigifySelectionColors)

    bpy.utils.unregister_class(RigifyArmatureLayer)
    bpy.utils.unregister_class(RigifyPreferences)

    metarig_menu.unregister()
    ui.unregister()



# gpl: author Yadoob
# -*- coding: utf-8 -*-

import bpy
from bpy.types import (
        Operator,
        Panel,
        )
from bpy.props import StringProperty
from .warning_messages_utils import (
        warning_messages,
        c_data_has_images,
        )


class TEXTURE_OT_patern_rename(Operator):
    bl_idname = "texture.patern_rename"
    bl_label = "Texture Renamer"
    bl_description = ("Replace the Texture names pattern with the attached Image ones\n"
                      "Works on all Textures (Including Brushes) \n \n"
                      "The First field - the name pattern to replace \n"
                      "The Second - searches for existing names \n")
    bl_options = {'REGISTER', 'UNDO'}

    def_name = "Texture"    # default name
    is_not_undo = False     # prevent drawing props on undo
    named = StringProperty(
                name="Search for name",
                default=def_name
                )

    @classmethod
    def poll(cls, context):
        return c_data_has_images()

    def draw(self, context):
        layout = self.layout
        if self.is_not_undo is True:
            box = layout.box()
            box.prop(self, "named", text="Name pattern", icon="SYNTAX_ON")
            layout.separator()

            box = layout.box()
            box.prop_search(self, "named", bpy.data, "textures")
        else:
            layout.label(text="**Only Undo is available**", icon="INFO")

    def invoke(self, context, event):
        self.is_not_undo = True
        return context.window_manager.invoke_props_dialog(self)

    def execute(self, context):
        errors = []     # collect texture names without images attached
        tex_count = 0   # check if there is textures at all

        for texture in bpy.data.textures:
            try:
                if texture and self.named in texture.name and texture.type in {"IMAGE"}:
                    tex_count += 1
                    textname = ""
                    img = (bpy.data.textures[texture.name].image if bpy.data.textures[texture.name] else None)
                    if not img:
                        errors.append(str(texture.name))
                    for word in img.name:
                        if word != ".":
                            textname = textname + word
                        else:
                            break
                    texture.name = textname
                if texture.type != "IMAGE":  # rename specific textures as clouds, environnement map,...
                    texture.name = texture.type.lower()
            except:
                continue

        if tex_count == 0:
            warning_messages(self, 'NO_TEX_RENAME')
        elif errors:
            warning_messages(self, 'TEX_RENAME_F', errors, 'TEX')

        # reset name to default
        self.named = self.def_name

        self.is_not_undo = False

        return {'FINISHED'}


class TEXTURE_PT_rename_panel(Panel):
    # Creates a Panel in the scene context of the properties editor
    bl_label = "Texture Rename"
    bl_idname = "SCENE_PT_layout"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "texture"

    def draw(self, context):
        layout = self.layout
        layout.operator("texture.patern_rename")


def register():
    bpy.utils.register_module(__name__)
    pass


def unregister():
    bpy.utils.unregister_module(__name__)
    pass


if __name__ == "__main__":
    register()



# -*- coding: utf-8 -*-

# Voronoi diagram calculator/ Delaunay triangulator
#
# - Voronoi Diagram Sweepline algorithm and C code by Steven Fortune,
#   1987, http://ect.bell-labs.com/who/sjf/
# - Python translation to file voronoi.py by Bill Simons, 2005, http://www.oxfish.com/
# - Additional changes for QGIS by Carson Farmer added November 2010
# - 2012 Ported to Python 3 and additional clip functions by domlysz at gmail.com
#
# Calculate Delaunay triangulation or the Voronoi polygons for a set of
# 2D input points.
#
# Derived from code bearing the following notice:
#
#  The author of this software is Steven Fortune.  Copyright (c) 1994 by AT&T
#  Bell Laboratories.
#  Permission to use, copy, modify, and distribute this software for any
#  purpose without fee is hereby granted, provided that this entire notice
#  is included in all copies of any software which is or includes a copy
#  or modification of this software and in all copies of the supporting
#  documentation for such software.
#  THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
#  WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR AT&T MAKE ANY
#  REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
#  OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
#
# Comments were incorporated from Shane O'Sullivan's translation of the
# original code into C++ (http://mapviewer.skynet.ie/voronoi.html)
#
# Steve Fortune's homepage: http://netlib.bell-labs.com/cm/cs/who/sjf/index.html
#
# For programmatic use, two functions are available:
#
#   computeVoronoiDiagram(points, xBuff, yBuff, polygonsOutput=False, formatOutput=False):
#   Takes :
#       - a list of point objects (which must have x and y fields).
#       - x and y buffer values which are the expansion percentages of the
#         bounding box rectangle including all input points.
#       Returns :
#       - With default options :
#         A list of 2-tuples, representing the two points of each Voronoi diagram edge.
#         Each point contains 2-tuples which are the x,y coordinates of point.
#         if formatOutput is True, returns :
#               - a list of 2-tuples, which are the x,y coordinates of the Voronoi diagram vertices.
#               - and a list of 2-tuples (v1, v2) representing edges of the Voronoi diagram.
#                 v1 and v2 are the indices of the vertices at the end of the edge.
#       - If polygonsOutput option is True, returns :
#         A dictionary of polygons, keys are the indices of the input points,
#         values contains n-tuples representing the n points of each Voronoi diagram polygon.
#         Each point contains 2-tuples which are the x,y coordinates of point.
#         if formatOutput is True, returns :
#               - A list of 2-tuples, which are the x,y coordinates of the Voronoi diagram vertices.
#               - and a dictionary of input points indices. Values contains n-tuples representing
#                 the n points of each Voronoi diagram polygon.
#                 Each tuple contains the vertex indices of the polygon vertices.
#
#   computeDelaunayTriangulation(points):
#       Takes a list of point objects (which must have x and y fields).
#       Returns a list of 3-tuples: the indices of the points that form a Delaunay triangle.

import bpy
import math

# Globals
TOLERANCE = 1e-9
BIG_FLOAT = 1e38


class Context(object):

    def __init__(self):
        self.doPrint = 0
        self.debug = 0

        # tuple (xmin, xmax, ymin, ymax)
        self.extent = ()
        self.triangulate = False
        # list of vertex 2-tuples: (x,y)
        self.vertices = []
        # equation of line 3-tuple (a b c), for the equation of the line a*x+b*y = c
        self.lines = []

        # edge 3-tuple: (line index, vertex 1 index, vertex 2 index)
        # if either vertex index is -1, the edge extends to infinity
        self.edges = []
        # 3-tuple of vertex indices
        self.triangles = []
        # a dict of site:[edges] pairs
        self.polygons = {}


# Clip functions #
    def getClipEdges(self):
        xmin, xmax, ymin, ymax = self.extent
        clipEdges = []
        for edge in self.edges:
            equation = self.lines[edge[0]]       # line equation
            if edge[1] != -1 and edge[2] != -1:  # finite line
                x1, y1 = self.vertices[edge[1]][0], self.vertices[edge[1]][1]
                x2, y2 = self.vertices[edge[2]][0], self.vertices[edge[2]][1]
                pt1, pt2 = (x1, y1), (x2, y2)
                inExtentP1, inExtentP2 = self.inExtent(x1, y1), self.inExtent(x2, y2)
                if inExtentP1 and inExtentP2:
                    clipEdges.append((pt1, pt2))
                elif inExtentP1 and not inExtentP2:
                    pt2 = self.clipLine(x1, y1, equation, leftDir=False)
                    clipEdges.append((pt1, pt2))
                elif not inExtentP1 and inExtentP2:
                    pt1 = self.clipLine(x2, y2, equation, leftDir=True)
                    clipEdges.append((pt1, pt2))
            else:  # infinite line
                if edge[1] != -1:
                    x1, y1 = self.vertices[edge[1]][0], self.vertices[edge[1]][1]
                    leftDir = False
                else:
                    x1, y1 = self.vertices[edge[2]][0], self.vertices[edge[2]][1]
                    leftDir = True
                if self.inExtent(x1, y1):
                    pt1 = (x1, y1)
                    pt2 = self.clipLine(x1, y1, equation, leftDir)
                    clipEdges.append((pt1, pt2))
        return clipEdges

    def getClipPolygons(self, closePoly):
        xmin, xmax, ymin, ymax = self.extent
        poly = {}
        for inPtsIdx, edges in self.polygons.items():
            clipEdges = []
            for edge in edges:
                equation = self.lines[edge[0]]       # line equation
                if edge[1] != -1 and edge[2] != -1:  # finite line
                    x1, y1 = self.vertices[edge[1]][0], self.vertices[edge[1]][1]
                    x2, y2 = self.vertices[edge[2]][0], self.vertices[edge[2]][1]
                    pt1, pt2 = (x1, y1), (x2, y2)
                    inExtentP1, inExtentP2 = self.inExtent(x1, y1), self.inExtent(x2, y2)
                    if inExtentP1 and inExtentP2:
                        clipEdges.append((pt1, pt2))
                    elif inExtentP1 and not inExtentP2:
                        pt2 = self.clipLine(x1, y1, equation, leftDir=False)
                        clipEdges.append((pt1, pt2))
                    elif not inExtentP1 and inExtentP2:
                        pt1 = self.clipLine(x2, y2, equation, leftDir=True)
                        clipEdges.append((pt1, pt2))
                else:  # infinite line
                    if edge[1] != -1:
                        x1, y1 = self.vertices[edge[1]][0], self.vertices[edge[1]][1]
                        leftDir = False
                    else:
                        x1, y1 = self.vertices[edge[2]][0], self.vertices[edge[2]][1]
                        leftDir = True
                    if self.inExtent(x1, y1):
                        pt1 = (x1, y1)
                        pt2 = self.clipLine(x1, y1, equation, leftDir)
                        clipEdges.append((pt1, pt2))
            # create polygon definition from edges and check if polygon is completely closed
            polyPts, complete = self.orderPts(clipEdges)
            if not complete:
                startPt = polyPts[0]
                endPt = polyPts[-1]
                # if start & end points are collinear then they are along an extent border
                if startPt[0] == endPt[0] or startPt[1] == endPt[1]:
                    polyPts.append(polyPts[0])  # simple close
                else:  # close at extent corner
                    # upper left
                    if (startPt[0] == xmin and endPt[1] == ymax) or (endPt[0] == xmin and startPt[1] == ymax):
                        polyPts.append((xmin, ymax))  # corner point
                        polyPts.append(polyPts[0])    # close polygon
                    # upper right
                    if (startPt[0] == xmax and endPt[1] == ymax) or (endPt[0] == xmax and startPt[1] == ymax):
                        polyPts.append((xmax, ymax))
                        polyPts.append(polyPts[0])
                    # bottom right
                    if (startPt[0] == xmax and endPt[1] == ymin) or (endPt[0] == xmax and startPt[1] == ymin):
                        polyPts.append((xmax, ymin))
                        polyPts.append(polyPts[0])
                    # bottom left
                    if (startPt[0] == xmin and endPt[1] == ymin) or (endPt[0] == xmin and startPt[1] == ymin):
                        polyPts.append((xmin, ymin))
                        polyPts.append(polyPts[0])
            if not closePoly:  # unclose polygon
                polyPts = polyPts[:-1]
            poly[inPtsIdx] = polyPts
        return poly

    def clipLine(self, x1, y1, equation, leftDir):
        xmin, xmax, ymin, ymax = self.extent
        a, b, c = equation
        if b == 0:       # vertical line
            if leftDir:  # left is bottom of vertical line
                return (x1, ymax)
            else:
                return (x1, ymin)
        elif a == 0:     # horizontal line
            if leftDir:
                return (xmin, y1)
            else:
                return (xmax, y1)
        else:
            y2_at_xmin = (c - a * xmin) / b
            y2_at_xmax = (c - a * xmax) / b
            x2_at_ymin = (c - b * ymin) / a
            x2_at_ymax = (c - b * ymax) / a
            intersectPts = []
            if ymin <= y2_at_xmin <= ymax:  # valid intersect point
                intersectPts.append((xmin, y2_at_xmin))
            if ymin <= y2_at_xmax <= ymax:
                intersectPts.append((xmax, y2_at_xmax))
            if xmin <= x2_at_ymin <= xmax:
                intersectPts.append((x2_at_ymin, ymin))
            if xmin <= x2_at_ymax <= xmax:
                intersectPts.append((x2_at_ymax, ymax))
            # delete duplicate (happens if intersect point is at extent corner)
            intersectPts = set(intersectPts)
            # choose target intersect point
            if leftDir:
                pt = min(intersectPts)  # smaller x value
            else:
                pt = max(intersectPts)
            return pt

    def inExtent(self, x, y):
        xmin, xmax, ymin, ymax = self.extent
        return x >= xmin and x <= xmax and y >= ymin and y <= ymax

    def orderPts(self, edges):
        poly = []  # returned polygon points list [pt1, pt2, pt3, pt4 ....]
        pts = []
        # get points list
        for edge in edges:
            pts.extend([pt for pt in edge])
        # try to get start & end point
        try:
            startPt, endPt = [pt for pt in pts if pts.count(pt) < 2]  # start and end point aren't duplicate
        except:  # all points are duplicate --> polygon is complete --> append some or other edge points
            complete = True
            firstIdx = 0
            poly.append(edges[0][0])
            poly.append(edges[0][1])
        else:  # incomplete --> append the first edge points
            complete = False
            # search first edge
            for i, edge in enumerate(edges):
                if startPt in edge:  # find
                    firstIdx = i
                    break
            poly.append(edges[firstIdx][0])
            poly.append(edges[firstIdx][1])
            if poly[0] != startPt:
                poly.reverse()
        # append next points in list
        del edges[firstIdx]
        while edges:  # all points will be treated when edges list will be empty
            currentPt = poly[-1]  # last item
            for i, edge in enumerate(edges):
                if currentPt == edge[0]:
                    poly.append(edge[1])
                    break
                elif currentPt == edge[1]:
                    poly.append(edge[0])
                    break
            del edges[i]
        return poly, complete

    def setClipBuffer(self, xpourcent, ypourcent):
        xmin, xmax, ymin, ymax = self.extent
        witdh = xmax - xmin
        height = ymax - ymin
        xmin = xmin - witdh * xpourcent / 100
        xmax = xmax + witdh * xpourcent / 100
        ymin = ymin - height * ypourcent / 100
        ymax = ymax + height * ypourcent / 100
        self.extent = xmin, xmax, ymin, ymax

    # End clip functions #

    def outSite(self, s):
        if(self.debug):
            print("site (%d) at %f %f" % (s.sitenum, s.x, s.y))
        elif(self.triangulate):
            pass
        elif(self.doPrint):
            print("s %f %f" % (s.x, s.y))

    def outVertex(self, s):
        self.vertices.append((s.x, s.y))
        if(self.debug):
            print("vertex(%d) at %f %f" % (s.sitenum, s.x, s.y))
        elif(self.triangulate):
            pass
        elif(self.doPrint):
            print("v %f %f" % (s.x, s.y))

    def outTriple(self, s1, s2, s3):
        self.triangles.append((s1.sitenum, s2.sitenum, s3.sitenum))
        if (self.debug):
            print("circle through left=%d right=%d bottom=%d" % (s1.sitenum, s2.sitenum, s3.sitenum))
        elif (self.triangulate and self.doPrint):
            print("%d %d %d" % (s1.sitenum, s2.sitenum, s3.sitenum))

    def outBisector(self, edge):
        self.lines.append((edge.a, edge.b, edge.c))
        if (self.debug):
            print("line(%d) %gx+%gy=%g, bisecting %d %d" % (edge.edgenum, edge.a, edge.b,
                                                            edge.c, edge.reg[0].sitenum,
                                                            edge.reg[1].sitenum)
                )
        elif(self.doPrint):
            print("l %f %f %f" % (edge.a, edge.b, edge.c))

    def outEdge(self, edge):
        sitenumL = -1
        if edge.ep[Edge.LE] is not None:
            sitenumL = edge.ep[Edge.LE].sitenum
        sitenumR = -1
        if edge.ep[Edge.RE] is not None:
            sitenumR = edge.ep[Edge.RE].sitenum

        # polygons dict add by CF
        if edge.reg[0].sitenum not in self.polygons:
            self.polygons[edge.reg[0].sitenum] = []
        if edge.reg[1].sitenum not in self.polygons:
            self.polygons[edge.reg[1].sitenum] = []
        self.polygons[edge.reg[0].sitenum].append((edge.edgenum, sitenumL, sitenumR))
        self.polygons[edge.reg[1].sitenum].append((edge.edgenum, sitenumL, sitenumR))

        self.edges.append((edge.edgenum, sitenumL, sitenumR))

        if (not self.triangulate):
            if (self.doPrint):
                print("e %d" % edge.edgenum)
                print(" %d " % sitenumL)
                print("%d" % sitenumR)


def voronoi(siteList, context):
    context.extent = siteList.extent
    edgeList = EdgeList(siteList.xmin, siteList.xmax, len(siteList))
    priorityQ = PriorityQueue(siteList.ymin, siteList.ymax, len(siteList))
    siteIter = siteList.iterator()

    bottomsite = siteIter.next()
    context.outSite(bottomsite)
    newsite = siteIter.next()
    minpt = Site(-BIG_FLOAT, -BIG_FLOAT)
    while True:
        if not priorityQ.isEmpty():
            minpt = priorityQ.getMinPt()

        if (newsite and (priorityQ.isEmpty() or newsite < minpt)):
            # newsite is smallest -  this is a site event
            context.outSite(newsite)

            # get first Halfedge to the LEFT and RIGHT of the new site
            lbnd = edgeList.leftbnd(newsite)
            rbnd = lbnd.right

            # if this halfedge has no edge, bot = bottom site (whatever that is)
            # create a new edge that bisects
            bot = lbnd.rightreg(bottomsite)
            edge = Edge.bisect(bot, newsite)
            context.outBisector(edge)

            # create a new Halfedge, setting its pm field to 0 and insert
            # this new bisector edge between the left and right vectors in
            # a linked list
            bisector = Halfedge(edge, Edge.LE)
            edgeList.insert(lbnd, bisector)

            # if the new bisector intersects with the left edge, remove
            # the left edge's vertex, and put in the new one
            p = lbnd.intersect(bisector)
            if p is not None:
                priorityQ.delete(lbnd)
                priorityQ.insert(lbnd, p, newsite.distance(p))

            # create a new Halfedge, setting its pm field to 1
            # insert the new Halfedge to the right of the original bisector
            lbnd = bisector
            bisector = Halfedge(edge, Edge.RE)
            edgeList.insert(lbnd, bisector)

            # if this new bisector intersects with the right Halfedge
            p = bisector.intersect(rbnd)
            if p is not None:
                # push the Halfedge into the ordered linked list of vertices
                priorityQ.insert(bisector, p, newsite.distance(p))

            newsite = siteIter.next()

        elif not priorityQ.isEmpty():
            # intersection is smallest - this is a vector (circle) event
            # pop the Halfedge with the lowest vector off the ordered list of
            # vectors.  Get the Halfedge to the left and right of the above HE
            # and also the Halfedge to the right of the right HE
            lbnd = priorityQ.popMinHalfedge()
            llbnd = lbnd.left
            rbnd = lbnd.right
            rrbnd = rbnd.right

            # get the Site to the left of the left HE and to the right of
            # the right HE which it bisects
            bot = lbnd.leftreg(bottomsite)
            top = rbnd.rightreg(bottomsite)

            # output the triple of sites, stating that a circle goes through them
            mid = lbnd.rightreg(bottomsite)
            context.outTriple(bot, top, mid)

            # get the vertex that caused this event and set the vertex number
            # couldn't do this earlier since we didn't know when it would be processed
            v = lbnd.vertex
            siteList.setSiteNumber(v)
            context.outVertex(v)

            # set the endpoint of the left and right Halfedge to be this vector
            if lbnd.edge.setEndpoint(lbnd.pm, v):
                context.outEdge(lbnd.edge)

            if rbnd.edge.setEndpoint(rbnd.pm, v):
                context.outEdge(rbnd.edge)

            # delete the lowest HE, remove all vertex events to do with the
            # right HE and delete the right HE
            edgeList.delete(lbnd)
            priorityQ.delete(rbnd)
            edgeList.delete(rbnd)

            # if the site to the left of the event is higher than the Site
            # to the right of it, then swap them and set 'pm' to RIGHT
            pm = Edge.LE
            if bot.y > top.y:
                bot, top = top, bot
                pm = Edge.RE

            # Create an Edge (or line) that is between the two Sites.  This
            # creates the formula of the line, and assigns a line number to it
            edge = Edge.bisect(bot, top)
            context.outBisector(edge)

            # create a HE from the edge
            bisector = Halfedge(edge, pm)

            # insert the new bisector to the right of the left HE
            # set one endpoint to the new edge to be the vector point 'v'
            # If the site to the left of this bisector is higher than the right
            # Site, then this endpoint is put in position 0; otherwise in pos 1
            edgeList.insert(llbnd, bisector)
            if edge.setEndpoint(Edge.RE - pm, v):
                context.outEdge(edge)

            # if left HE and the new bisector don't intersect, then delete
            # the left HE, and reinsert it
            p = llbnd.intersect(bisector)
            if p is not None:
                priorityQ.delete(llbnd)
                priorityQ.insert(llbnd, p, bot.distance(p))

            # if right HE and the new bisector don't intersect, then reinsert it
            p = bisector.intersect(rrbnd)
            if p is not None:
                priorityQ.insert(bisector, p, bot.distance(p))
        else:
            break

    he = edgeList.leftend.right
    while he is not edgeList.rightend:
        context.outEdge(he.edge)
        he = he.right
    Edge.EDGE_NUM = 0  # CF


def isEqual(a, b, relativeError=TOLERANCE):
    # is nearly equal to within the allowed relative error
    norm = max(abs(a), abs(b))
    return (norm < relativeError) or (abs(a - b) < (relativeError * norm))


class Site(object):

    def __init__(self, x=0.0, y=0.0, sitenum=0):
        self.x = x
        self.y = y
        self.sitenum = sitenum

    def dump(self):
        print("Site #%d (%g, %g)" % (self.sitenum, self.x, self.y))

    def __lt__(self, other):
        if self.y < other.y:
            return True
        elif self.y > other.y:
            return False
        elif self.x < other.x:
            return True
        elif self.x > other.x:
            return False
        else:
            return False

    def __eq__(self, other):
        if self.y == other.y and self.x == other.x:
            return True

    def distance(self, other):
        dx = self.x - other.x
        dy = self.y - other.y
        return math.sqrt(dx * dx + dy * dy)


class Edge(object):
    LE = 0  # left end indice --> edge.ep[Edge.LE]
    RE = 1  # right end indice
    EDGE_NUM = 0
    DELETED = {}  # marker value

    def __init__(self):
        self.a = 0.0  # equation of the line a*x+b*y = c
        self.b = 0.0
        self.c = 0.0
        self.ep = [None, None]  # end point (2 tuples of site)
        self.reg = [None, None]
        self.edgenum = 0

    def dump(self):
        print("(#%d a=%g, b=%g, c=%g)" % (self.edgenum, self.a, self.b, self.c))
        print("ep", self.ep)
        print("reg", self.reg)

    def setEndpoint(self, lrFlag, site):
        self.ep[lrFlag] = site
        if self.ep[Edge.RE - lrFlag] is None:
            return False
        return True

    @staticmethod
    def bisect(s1, s2):
        newedge = Edge()
        newedge.reg[0] = s1  # store the sites that this edge is bisecting
        newedge.reg[1] = s2

        # to begin with, there are no endpoints on the bisector - it goes to infinity
        # ep[0] and ep[1] are None

        # get the difference in x dist between the sites
        dx = float(s2.x - s1.x)
        dy = float(s2.y - s1.y)
        adx = abs(dx)  # make sure that the difference in positive
        ady = abs(dy)

        # get the slope of the line
        newedge.c = float(s1.x * dx + s1.y * dy + (dx * dx + dy * dy) * 0.5)
        if adx > ady:
            # set formula of line, with x fixed to 1
            newedge.a = 1.0
            newedge.b = dy / dx
            newedge.c /= dx
        else:
            # set formula of line, with y fixed to 1
            newedge.b = 1.0
            newedge.a = dx / dy
            newedge.c /= dy

        newedge.edgenum = Edge.EDGE_NUM
        Edge.EDGE_NUM += 1
        return newedge


class Halfedge(object):

    def __init__(self, edge=None, pm=Edge.LE):
        self.left = None    # left Halfedge in the edge list
        self.right = None   # right Halfedge in the edge list
        self.qnext = None   # priority queue linked list pointer
        self.edge = edge    # edge list Edge
        self.pm = pm
        self.vertex = None  # Site()
        self.ystar = BIG_FLOAT

    def dump(self):
        print("Halfedge--------------------------")
        print("left: ", self.left)
        print("right: ", self.right)
        print("edge: ", self.edge)
        print("pm: ", self.pm)
        print("vertex: "),
        if self.vertex:
            self.vertex.dump()
        else:
            print("None")
        print("ystar: ", self.ystar)

    def __lt__(self, other):
        if self.ystar < other.ystar:
            return True
        elif self.ystar > other.ystar:
            return False
        elif self.vertex.x < other.vertex.x:
            return True
        elif self.vertex.x > other.vertex.x:
            return False
        else:
            return False

    def __eq__(self, other):
        if self.ystar == other.ystar and self.vertex.x == other.vertex.x:
            return True

    def leftreg(self, default):
        if not self.edge:
            return default
        elif self.pm == Edge.LE:
            return self.edge.reg[Edge.LE]
        else:
            return self.edge.reg[Edge.RE]

    def rightreg(self, default):
        if not self.edge:
            return default
        elif self.pm == Edge.LE:
            return self.edge.reg[Edge.RE]
        else:
            return self.edge.reg[Edge.LE]

    # returns True if p is to right of halfedge self
    def isPointRightOf(self, pt):
        e = self.edge
        topsite = e.reg[1]
        right_of_site = pt.x > topsite.x

        if(right_of_site and self.pm == Edge.LE):
            return True

        if(not right_of_site and self.pm == Edge.RE):
            return False

        if(e.a == 1.0):
            dyp = pt.y - topsite.y
            dxp = pt.x - topsite.x
            fast = 0
            if ((not right_of_site and e.b < 0.0) or (right_of_site and e.b >= 0.0)):
                above = dyp >= e.b * dxp
                fast = above
            else:
                above = pt.x + pt.y * e.b > e.c
                if(e.b < 0.0):
                    above = not above
                if (not above):
                    fast = 1
            if (not fast):
                dxs = topsite.x - (e.reg[0]).x
                above = e.b * (dxp * dxp - dyp * dyp) < dxs * dyp * (1.0 + 2.0 * dxp / dxs + e.b * e.b)
                if(e.b < 0.0):
                    above = not above
        else:  # e.b == 1.0
            yl = e.c - e.a * pt.x
            t1 = pt.y - yl
            t2 = pt.x - topsite.x
            t3 = yl - topsite.y
            above = t1 * t1 > t2 * t2 + t3 * t3

        if(self.pm == Edge.LE):
            return above
        else:
            return not above

    # create a new site where the Halfedges el1 and el2 intersect
    def intersect(self, other):
        e1 = self.edge
        e2 = other.edge
        if (e1 is None) or (e2 is None):
            return None

        # if the two edges bisect the same parent return None
        if e1.reg[1] is e2.reg[1]:
            return None

        d = e1.a * e2.b - e1.b * e2.a
        if isEqual(d, 0.0):
            return None

        xint = (e1.c * e2.b - e2.c * e1.b) / d
        yint = (e2.c * e1.a - e1.c * e2.a) / d
        if e1.reg[1] < e2.reg[1]:
            he = self
            e = e1
        else:
            he = other
            e = e2

        rightOfSite = xint >= e.reg[1].x
        if((rightOfSite and he.pm == Edge.LE) or
                (not rightOfSite and he.pm == Edge.RE)):
            return None

        # create a new site at the point of intersection - this is a new
        # vector event waiting to happen
        return Site(xint, yint)


class EdgeList(object):

    def __init__(self, xmin, xmax, nsites):
        if xmin > xmax:
            xmin, xmax = xmax, xmin
        self.hashsize = int(2 * math.sqrt(nsites + 4))

        self.xmin = xmin
        self.deltax = float(xmax - xmin)
        self.hash = [None] * self.hashsize

        self.leftend = Halfedge()
        self.rightend = Halfedge()
        self.leftend.right = self.rightend
        self.rightend.left = self.leftend
        self.hash[0] = self.leftend
        self.hash[-1] = self.rightend

    def insert(self, left, he):
        he.left = left
        he.right = left.right
        left.right.left = he
        left.right = he

    def delete(self, he):
        he.left.right = he.right
        he.right.left = he.left
        he.edge = Edge.DELETED

    # Get entry from hash table, pruning any deleted nodes
    def gethash(self, b):
        if(b < 0 or b >= self.hashsize):
            return None
        he = self.hash[b]
        if he is None or he.edge is not Edge.DELETED:
            return he

        #  Hash table points to deleted half edge.  Patch as necessary.
        self.hash[b] = None
        return None

    def leftbnd(self, pt):
        # Use hash table to get close to desired halfedge
        bucket = int(((pt.x - self.xmin) / self.deltax * self.hashsize))

        if(bucket < 0):
            bucket = 0

        if(bucket >= self.hashsize):
            bucket = self.hashsize - 1

        he = self.gethash(bucket)
        if(he is None):
            i = 1
            while True:
                he = self.gethash(bucket - i)
                if (he is not None):
                    break
                he = self.gethash(bucket + i)
                if (he is not None):
                    break
                i += 1

        # Now search linear list of halfedges for the corect one
        if (he is self.leftend) or (he is not self.rightend and he.isPointRightOf(pt)):
            he = he.right
            while he is not self.rightend and he.isPointRightOf(pt):
                he = he.right
            he = he.left
        else:
            he = he.left
            while (he is not self.leftend and not he.isPointRightOf(pt)):
                he = he.left

        # Update hash table and reference counts
        if(bucket > 0 and bucket < self.hashsize - 1):
            self.hash[bucket] = he
        return he


class PriorityQueue(object):

    def __init__(self, ymin, ymax, nsites):
        self.ymin = ymin
        self.deltay = ymax - ymin
        self.hashsize = int(4 * math.sqrt(nsites))
        self.count = 0
        self.minidx = 0
        self.hash = []
        for i in range(self.hashsize):
            self.hash.append(Halfedge())

    def __len__(self):
        return self.count

    def isEmpty(self):
        return self.count == 0

    def insert(self, he, site, offset):
        he.vertex = site
        he.ystar = site.y + offset
        last = self.hash[self.getBucket(he)]
        next = last.qnext
        while((next is not None) and he > next):
            last = next
            next = last.qnext
        he.qnext = last.qnext
        last.qnext = he
        self.count += 1

    def delete(self, he):
        if (he.vertex is not None):
            last = self.hash[self.getBucket(he)]
            while last.qnext is not he:
                last = last.qnext
            last.qnext = he.qnext
            self.count -= 1
            he.vertex = None

    def getBucket(self, he):
        bucket = int(((he.ystar - self.ymin) / self.deltay) * self.hashsize)
        if bucket < 0:
            bucket = 0
        if bucket >= self.hashsize:
            bucket = self.hashsize - 1
        if bucket < self.minidx:
            self.minidx = bucket
        return bucket

    def getMinPt(self):
        while(self.hash[self.minidx].qnext is None):
            self.minidx += 1
        he = self.hash[self.minidx].qnext
        x = he.vertex.x
        y = he.ystar
        return Site(x, y)

    def popMinHalfedge(self):
        curr = self.hash[self.minidx].qnext
        self.hash[self.minidx].qnext = curr.qnext
        self.count -= 1
        return curr


class SiteList(object):

    def __init__(self, pointList):
        self.__sites = []
        self.__sitenum = 0

        self.__xmin = min([pt.x for pt in pointList])
        self.__ymin = min([pt.y for pt in pointList])
        self.__xmax = max([pt.x for pt in pointList])
        self.__ymax = max([pt.y for pt in pointList])
        self.__extent = (self.__xmin, self.__xmax, self.__ymin, self.__ymax)

        for i, pt in enumerate(pointList):
            self.__sites.append(Site(pt.x, pt.y, i))
        self.__sites.sort()

    def setSiteNumber(self, site):
        site.sitenum = self.__sitenum
        self.__sitenum += 1

    class Iterator(object):

        def __init__(this, lst):
            this.generator = (s for s in lst)

        def __iter__(this):
            return this

        def next(this):
            try:
                # Note: Blender is Python 3.x so no need for 2.x checks
                return this.generator.__next__()
            except StopIteration:
                return None

    def iterator(self):
        return SiteList.Iterator(self.__sites)

    def __iter__(self):
        return SiteList.Iterator(self.__sites)

    def __len__(self):
        return len(self.__sites)

    def _getxmin(self):
        return self.__xmin

    def _getymin(self):
        return self.__ymin

    def _getxmax(self):
        return self.__xmax

    def _getymax(self):
        return self.__ymax

    def _getextent(self):
        return self.__extent

    xmin = property(_getxmin)
    ymin = property(_getymin)
    xmax = property(_getxmax)
    ymax = property(_getymax)
    extent = property(_getextent)


def computeVoronoiDiagram(points, xBuff=0, yBuff=0, polygonsOutput=False,
                          formatOutput=False, closePoly=True):
    """
    Takes :
    - a list of point objects (which must have x and y fields).
    - x and y buffer values which are the expansion percentages of the bounding box
        rectangle including all input points.
    Returns :
    - With default options :
      A list of 2-tuples, representing the two points of each Voronoi diagram edge.
      Each point contains 2-tuples which are the x,y coordinates of point.
      if formatOutput is True, returns :
                    - a list of 2-tuples, which are the x,y coordinates of the Voronoi diagram vertices.
                    - and a list of 2-tuples (v1, v2) representing edges of the Voronoi diagram.
                      v1 and v2 are the indices of the vertices at the end of the edge.
    - If polygonsOutput option is True, returns :
      A dictionary of polygons, keys are the indices of the input points,
      values contains n-tuples representing the n points of each Voronoi diagram polygon.
      Each point contains 2-tuples which are the x,y coordinates of point.
      if formatOutput is True, returns :
                    - A list of 2-tuples, which are the x,y coordinates of the Voronoi diagram vertices.
                    - and a dictionary of input points indices. Values contains n-tuples representing
                      the n points of each Voronoi diagram polygon.
                      Each tuple contains the vertex indices of the polygon vertices.
    - if closePoly is True then, in the list of points of a polygon, last point will be the same of first point
    """
    siteList = SiteList(points)
    context = Context()
    voronoi(siteList, context)
    context.setClipBuffer(xBuff, yBuff)
    if not polygonsOutput:
        clipEdges = context.getClipEdges()
        if formatOutput:
            vertices, edgesIdx = formatEdgesOutput(clipEdges)
            return vertices, edgesIdx
        else:
            return clipEdges
    else:
        clipPolygons = context.getClipPolygons(closePoly)
        if formatOutput:
            vertices, polyIdx = formatPolygonsOutput(clipPolygons)
            return vertices, polyIdx
        else:
            return clipPolygons


def formatEdgesOutput(edges):
    # get list of points
    pts = []
    for edge in edges:
        pts.extend(edge)
    # get unique values
    pts = set(pts)  # unique values (tuples are hashable)
    # get dict {values:index}
    valuesIdxDict = dict(zip(pts, range(len(pts))))
    # get edges index reference
    edgesIdx = []
    for edge in edges:
        edgesIdx.append([valuesIdxDict[pt] for pt in edge])
    return list(pts), edgesIdx


def formatPolygonsOutput(polygons):
    # get list of points
    pts = []
    for poly in polygons.values():
        pts.extend(poly)
    # get unique values
    pts = set(pts)  # unique values (tuples are hashable)
    # get dict {values:index}
    valuesIdxDict = dict(zip(pts, range(len(pts))))
    # get polygons index reference
    polygonsIdx = {}
    for inPtsIdx, poly in polygons.items():
        polygonsIdx[inPtsIdx] = [valuesIdxDict[pt] for pt in poly]
    return list(pts), polygonsIdx


def computeDelaunayTriangulation(points):
    """ Takes a list of point objects (which must have x and y fields).
            Returns a list of 3-tuples: the indices of the points that form a
            Delaunay triangle.
    """
    siteList = SiteList(points)
    context = Context()
    context.triangulate = True
    voronoi(siteList, context)
    return context.triangles



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

bl_info = {
    "name": "Geodesic Domes2",
    "author": "Noctumsolis, PKHG, Meta Androcto, Andy Houston",
    "version": (0, 3, 2),
    "blender": (2, 7, 1),
    "location": "Toolshelf > Create Tab",
    "description": "Create geodesic dome type objects.",
    "warning": "",
    "wiki_url": "https://wiki.blender.org/index.php/Extensions:2.6/Py/Scripts/Modeling/Geodesic_Domes",
    "tracker_url": "",
    "category": "Mesh"}

if "bpy" in locals():
    import importlib
    importlib.reload(third_domes_panel_271)

else:
    from . import third_domes_panel_271

import bpy


def register():
    bpy.utils.register_module(__name__)


def unregister():
    bpy.utils.unregister_module(__name__)


if __name__ == "__main__":
    register()



# gpl author: Stanislav Blinov

bl_info = {
    "name": "V/E/F Context Menu",
    "author": "Stanislav Blinov",
    "version": (1, 0, 1),
    "blender": (2, 78, 0),
    "description": "Vert Edge Face Double Right Click Edit Mode",
    "category": "Mesh",
}

import bpy
import bpy_extras
from bpy.types import (
        Menu,
        Operator,
        )


class MESH_MT_CombinedMenu(Menu):
    bl_idname = "mesh.addon_combined_component_menu"
    bl_label = "Components"

    @classmethod
    def poll(cls, context):
        return context.mode == 'EDIT_MESH'

    def draw(self, context):
        layout = self.layout

        mode = context.tool_settings.mesh_select_mode
        if mode[0]:
            layout.menu("VIEW3D_MT_edit_mesh_vertices")
        if mode[1]:
            layout.menu("VIEW3D_MT_edit_mesh_edges")
        if mode[2]:
            layout.menu("VIEW3D_MT_edit_mesh_faces")


class MESH_OT_CallContextMenu(Operator):
    bl_idname = "mesh.addon_call_context_menu"
    bl_label = "Context Menu"

    @classmethod
    def poll(cls, context):
        return context.mode == 'EDIT_MESH'

    def execute(self, context):
        mode = context.tool_settings.mesh_select_mode
        num = sum(int(m) for m in mode)
        if num == 1:
            if mode[0]:
                return bpy.ops.wm.call_menu(name="VIEW3D_MT_edit_mesh_vertices")
            if mode[1]:
                return bpy.ops.wm.call_menu(name="VIEW3D_MT_edit_mesh_edges")
            if mode[2]:
                return bpy.ops.wm.call_menu(name="VIEW3D_MT_edit_mesh_faces")
        else:
            return bpy.ops.wm.call_menu(name=MESH_MT_CombinedMenu.bl_idname)


classes = (
    MESH_MT_CombinedMenu,
    MESH_OT_CallContextMenu,
    )


KEYMAPS = (
    # First, keymap identifiers (last bool is True for modal km).
    (("3D View", "VIEW_3D", "WINDOW", False), (
    # Then a tuple of keymap items, defined by a dict of kwargs
    # for the km new func, and a tuple of tuples (name, val)
    # for ops properties, if needing non-default values.
        ({"idname": MESH_OT_CallContextMenu.bl_idname, "type": 'RIGHTMOUSE', "value": 'DOUBLE_CLICK'},
         ()),
    )),
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    bpy_extras.keyconfig_utils.addon_keymap_register(bpy.context.window_manager, KEYMAPS)


def unregister():
    bpy_extras.keyconfig_utils.addon_keymap_unregister(bpy.context.window_manager, KEYMAPS)

    for cls in classes:
        bpy.utils.unregister_class(cls)


if __name__ == "__main__":
    register()



import os

TAB_SIZE = 4
LIN_SIZE = 120

IGNORE = (

    # particles
    "source/blender/blenkernel/intern/boids.c",
    "source/blender/blenkernel/intern/cloth.c",
    "source/blender/blenkernel/intern/collision.c",
    "source/blender/blenkernel/intern/effect.c",
    "source/blender/blenkernel/intern/implicit.c",
    "source/blender/blenkernel/intern/particle.c",
    "source/blender/blenkernel/intern/particle_system.c",
    "source/blender/blenkernel/intern/pointcache.c",
    "source/blender/blenkernel/intern/sca.c",
    "source/blender/blenkernel/intern/softbody.c",
    "source/blender/blenkernel/intern/smoke.c",

    "source/blender/blenlib/intern/fnmatch.c",
    "source/blender/blenlib/intern/md5.c",
    "source/blender/blenlib/intern/voxel.c",

    "source/blender/blenloader/intern/readfile.c",
    "source/blender/blenloader/intern/versioning_250.c",
    "source/blender/blenloader/intern/versioning_legacy.c",

    "source/blender/editors/space_logic/logic_buttons.c",
    "source/blender/editors/space_logic/logic_window.c",

    "source/blender/imbuf/intern/dds/DirectDrawSurface.cpp",

    "source/blender/opencl/intern/clew.c",
    "source/blender/opencl/intern/clew.h",
)

IGNORE_DIR = (
    "source/blender/collada",
    "source/blender/render",
    "source/blender/editors/physics",
    "source/blender/editors/space_logic",
    "source/blender/freestyle",
)


SOURCE_DIR = os.path.normpath(os.path.abspath(os.path.normpath(
    os.path.join(os.path.dirname(__file__), "..", "..", ".."))))



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "4.4"
__date__ = "2 Aug 2017"

from bpy.props import (
        BoolProperty,
        FloatProperty,
        FloatVectorProperty,
        )
from bpy.types import AddonPreferences


class MUV_Preferences(AddonPreferences):
    """Preferences class: Preferences for this add-on"""

    bl_idname = __package__

    # enable/disable switcher
    enable_texproj = BoolProperty(
        name="Texture Projection",
        default=True)
    enable_uvbb = BoolProperty(
        name="Bounding Box",
        default=True)

    # for Texture Projection
    texproj_canvas_padding = FloatVectorProperty(
        name="Canvas Padding",
        description="Canvas Padding",
        size=2,
        max=50.0,
        min=0.0,
        default=(20.0, 20.0))

    # for UV Bounding Box
    uvbb_cp_size = FloatProperty(
        name="Size",
        description="Control Point Size",
        default=6.0,
        min=3.0,
        max=100.0)
    uvbb_cp_react_size = FloatProperty(
        name="React Size",
        description="Size event fired",
        default=10.0,
        min=3.0,
        max=100.0)

    def draw(self, _):
        layout = self.layout

        layout.label("Switch Enable/Disable and Configurate Features:")

        layout.prop(self, "enable_texproj")
        if self.enable_texproj:
            sp = layout.split(percentage=0.05)
            col = sp.column()       # spacer
            sp = sp.split(percentage=0.3)
            col = sp.column()
            col.label("Texture Display: ")
            col.prop(self, "texproj_canvas_padding")

        layout.prop(self, "enable_uvbb")
        if self.enable_uvbb:
            sp = layout.split(percentage=0.05)
            col = sp.column()       # spacer
            sp = sp.split(percentage=0.3)
            col = sp.column()
            col.label("Control Point: ")
            col.prop(self, "uvbb_cp_size")
            col.prop(self, "uvbb_cp_react_size")

        layout.label("Description:")
        column = layout.column(align=True)
        column.label("Magic UV is composed of many UV editing features.")
        column.label("See tutorial page if you are new to this add-on.")
        column.label("https://github.com/nutti/Magic-UV/wiki/Tutorial")

        layout.label("Location:")

        row = layout.row(align=True)
        sp = row.split(percentage=0.3)
        sp.label("View3D > U")
        sp = sp.split(percentage=1.0)
        col = sp.column(align=True)
        col.label("Copy/Paste UV Coordinates")
        col.label("Copy/Paste UV Coordinates (by selection sequence)")
        col.label("Flip/Rotate UVs")
        col.label("Transfer UV")
        col.label("Move UV from 3D View")
        col.label("Texture Lock")
        col.label("Mirror UV")
        col.label("World Scale UV")
        col.label("Unwrap Constraint")
        col.label("Preserve UV Aspect")

        row = layout.row(align=True)
        sp = row.split(percentage=0.3)
        sp.label("View3D > Object")
        sp = sp.split(percentage=1.0)
        col = sp.column(align=True)
        col.label("Copy/Paste UV Coordinates (Among same objects)")

        row = layout.row(align=True)
        sp = row.split(percentage=0.3)
        sp.label("ImageEditor > Property Panel")
        sp = sp.split(percentage=1.0)
        col = sp.column(align=True)
        col.label("Manipulate UV with Bounding Box in UV Editor")

        row = layout.row(align=True)
        sp = row.split(percentage=0.3)
        sp.label("View3D > Property Panel")
        sp = sp.split(percentage=1.0)
        col = sp.column(align=True)
        col.label("Texture Projection")

        row = layout.row(align=True)
        sp = row.split(percentage=0.3)
        sp.label("ImageEditor > UVs")
        sp = sp.split(percentage=1.0)
        col = sp.column(align=True)
        col.label("Pack UV (with same UV island packing)")



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "4.4"
__date__ = "2 Aug 2017"

import bpy
import bmesh
from bpy.props import (
        StringProperty,
        BoolProperty,
        IntProperty,
        EnumProperty,
        )
from . import muv_common


class MUV_CPUVSelSeqCopyUV(bpy.types.Operator):
    """
    Operation class: Copy UV coordinate by selection sequence
    """

    bl_idname = "uv.muv_cpuv_selseq_copy_uv"
    bl_label = "Copy UV (Selection Sequence) (Operation)"
    bl_description = "Copy UV data by selection sequence (Operation)"
    bl_options = {'REGISTER', 'UNDO'}

    uv_map = StringProperty(options={'HIDDEN'})

    def execute(self, context):
        props = context.scene.muv_props.cpuv_selseq
        if self.uv_map == "":
            self.report({'INFO'}, "Copy UV coordinate (selection sequence)")
        else:
            self.report(
                {'INFO'},
                "Copy UV coordinate (selection sequence) (UV map:%s)"
                % (self.uv_map))
        obj = context.active_object
        bm = bmesh.from_edit_mesh(obj.data)
        if muv_common.check_version(2, 73, 0) >= 0:
            bm.faces.ensure_lookup_table()

        # get UV layer
        if self.uv_map == "":
            if not bm.loops.layers.uv:
                self.report(
                    {'WARNING'}, "Object must have more than one UV map")
                return {'CANCELLED'}
            uv_layer = bm.loops.layers.uv.verify()
        else:
            uv_layer = bm.loops.layers.uv[self.uv_map]

        # get selected face
        props.src_uvs = []
        props.src_pin_uvs = []
        props.src_seams = []
        for hist in bm.select_history:
            if isinstance(hist, bmesh.types.BMFace) and hist.select:
                uvs = [l[uv_layer].uv.copy() for l in hist.loops]
                pin_uvs = [l[uv_layer].pin_uv for l in hist.loops]
                seams = [l.edge.seam  for l in hist.loops]
                props.src_uvs.append(uvs)
                props.src_pin_uvs.append(pin_uvs)
                props.src_seams.append(seams)
        if len(props.src_uvs) == 0 or len(props.src_pin_uvs) == 0:
            self.report({'WARNING'}, "No faces are selected")
            return {'CANCELLED'}
        self.report({'INFO'}, "%d face(s) are selected" % len(props.src_uvs))

        return {'FINISHED'}


class MUV_CPUVSelSeqCopyUVMenu(bpy.types.Menu):
    """
    Menu class: Copy UV coordinate by selection sequence
    """

    bl_idname = "uv.muv_cpuv_selseq_copy_uv_menu"
    bl_label = "Copy UV (Selection Sequence)"
    bl_description = "Copy UV coordinate by selection sequence"

    def draw(self, context):
        layout = self.layout
        obj = context.active_object
        bm = bmesh.from_edit_mesh(obj.data)
        uv_maps = bm.loops.layers.uv.keys()
        layout.operator(
            MUV_CPUVSelSeqCopyUV.bl_idname,
            text="[Default]", icon="IMAGE_COL").uv_map = ""
        for m in uv_maps:
            layout.operator(
                MUV_CPUVSelSeqCopyUV.bl_idname,
                text=m, icon="IMAGE_COL").uv_map = m


class MUV_CPUVSelSeqPasteUV(bpy.types.Operator):
    """
    Operation class: Paste UV coordinate by selection sequence
    """

    bl_idname = "uv.muv_cpuv_selseq_paste_uv"
    bl_label = "Paste UV (Selection Sequence) (Operation)"
    bl_description = "Paste UV coordinate by selection sequence (Operation)"
    bl_options = {'REGISTER', 'UNDO'}

    uv_map = StringProperty(options={'HIDDEN'})
    strategy = EnumProperty(
        name="Strategy",
        description="Paste Strategy",
        items=[
            ('N_N', 'N:N', 'Number of faces must be equal to source'),
            ('N_M', 'N:M', 'Number of faces must not be equal to source')
        ],
        default="N_M"
    )
    flip_copied_uv = BoolProperty(
        name="Flip Copied UV",
        description="Flip Copied UV...",
        default=False
    )
    rotate_copied_uv = IntProperty(
        default=0,
        name="Rotate Copied UV",
        min=0,
        max=30
    )
    copy_seams = BoolProperty(
        name="Copy Seams",
        description="Copy Seams",
        default=True
    )

    def execute(self, context):
        props = context.scene.muv_props.cpuv_selseq
        if len(props.src_uvs) == 0 or len(props.src_pin_uvs) == 0:
            self.report({'WARNING'}, "Need copy UV at first")
            return {'CANCELLED'}
        if self.uv_map == "":
            self.report({'INFO'}, "Paste UV coordinate (selection sequence)")
        else:
            self.report(
                {'INFO'},
                "Paste UV coordinate (selection sequence) (UV map:%s)"
                % (self.uv_map))

        obj = context.active_object
        bm = bmesh.from_edit_mesh(obj.data)
        if muv_common.check_version(2, 73, 0) >= 0:
            bm.faces.ensure_lookup_table()

        # get UV layer
        if self.uv_map == "":
            if not bm.loops.layers.uv:
                self.report(
                    {'WARNING'}, "Object must have more than one UV map")
                return {'CANCELLED'}
            uv_layer = bm.loops.layers.uv.verify()
        else:
            uv_layer = bm.loops.layers.uv[self.uv_map]

        # get selected face
        dest_uvs = []
        dest_pin_uvs = []
        dest_seams = []
        dest_face_indices = []
        for hist in bm.select_history:
            if isinstance(hist, bmesh.types.BMFace) and hist.select:
                dest_face_indices.append(hist.index)
                uvs = [l[uv_layer].uv.copy() for l in hist.loops]
                pin_uvs = [l[uv_layer].pin_uv for l in hist.loops]
                seams = [l.edge.seam for l in hist.loops]
                dest_uvs.append(uvs)
                dest_pin_uvs.append(pin_uvs)
                dest_seams.append(seams)
        if len(dest_uvs) == 0 or len(dest_pin_uvs) == 0:
            self.report({'WARNING'}, "No faces are selected")
            return {'CANCELLED'}
        if self.strategy == 'N_N' and len(props.src_uvs) != len(dest_uvs):
            self.report(
                {'WARNING'},
                "Number of selected faces is different from copied faces " +
                "(src:%d, dest:%d)"
                % (len(props.src_uvs), len(dest_uvs)))
            return {'CANCELLED'}

        # paste
        for i, idx in enumerate(dest_face_indices):
            suv = None
            spuv = None
            ss = None
            duv = None
            if self.strategy == 'N_N':
                suv = props.src_uvs[i]
                spuv = props.src_pin_uvs[i]
                ss = props.src_seams[i]
                duv = dest_uvs[i]
            elif self.strategy == 'N_M':
                suv = props.src_uvs[i % len(props.src_uvs)]
                spuv = props.src_pin_uvs[i % len(props.src_pin_uvs)]
                ss = props.src_seams[i % len(props.src_seams)]
                duv = dest_uvs[i]
            if len(suv) != len(duv):
                self.report({'WARNING'}, "Some faces are different size")
                return {'CANCELLED'}
            suvs_fr = [uv for uv in suv]
            spuvs_fr = [pin_uv for pin_uv in spuv]
            ss_fr = [s for s in ss]
            # flip UVs
            if self.flip_copied_uv is True:
                suvs_fr.reverse()
                spuvs_fr.reverse()
                ss_fr.reverse()
            # rotate UVs
            for _ in range(self.rotate_copied_uv):
                uv = suvs_fr.pop()
                pin_uv = spuvs_fr.pop()
                s = ss_fr.pop()
                suvs_fr.insert(0, uv)
                spuvs_fr.insert(0, pin_uv)
                ss_fr.insert(0, s)
            # paste UVs
            for l, suv, spuv, ss in zip(bm.faces[idx].loops, suvs_fr, spuvs_fr, ss_fr):
                l[uv_layer].uv = suv
                l[uv_layer].pin_uv = spuv
                if self.copy_seams is True:
                    l.edge.seam = ss

        self.report({'INFO'}, "%d face(s) are copied" % len(dest_uvs))

        bmesh.update_edit_mesh(obj.data)
        if self.copy_seams is True:
            obj.data.show_edge_seams = True

        return {'FINISHED'}


class MUV_CPUVSelSeqPasteUVMenu(bpy.types.Menu):
    """
    Menu class: Paste UV coordinate by selection sequence
    """

    bl_idname = "uv.muv_cpuv_selseq_paste_uv_menu"
    bl_label = "Paste UV (Selection Sequence)"
    bl_description = "Paste UV coordinate by selection sequence"

    def draw(self, context):
        layout = self.layout
        # create sub menu
        obj = context.active_object
        bm = bmesh.from_edit_mesh(obj.data)
        uv_maps = bm.loops.layers.uv.keys()
        layout.operator(
            MUV_CPUVSelSeqPasteUV.bl_idname,
            text="[Default]", icon="IMAGE_COL").uv_map = ""
        for m in uv_maps:
            layout.operator(
                MUV_CPUVSelSeqPasteUV.bl_idname,
                text=m, icon="IMAGE_COL").uv_map = m



# gpl: author Dannyboy

bl_info = {
    "name": "Add Random Box Structure",
    "author": "Dannyboy",
    "version": (1, 0, 1),
    "location": "View3D > Add > Make Box Structure",
    "description": "Fill selected box shaped meshes with randomly sized cubes",
    "warning": "",
    "wiki_url": "",
    "tracker_url": "dannyboypython.blogspot.com",
    "category": "Object"}

import bpy
import random
from bpy.types import Operator
from bpy.props import (
        BoolProperty,
        FloatProperty,
        FloatVectorProperty,
        IntProperty,
        )


class makestructure(Operator):
    bl_idname = "object.make_structure"
    bl_label = "Add Random Box Structure"
    bl_description = ("Create a randomized structure made of boxes\n"
                      "with various control parameters\n"
                      "Needs an existing Active Mesh Object")
    bl_options = {'REGISTER', 'UNDO'}

    dc = BoolProperty(
            name="Delete Base Mesh(es)",
            default=True
            )
    wh = BoolProperty(
            name="Stay Within Bounds",
            description="Keeps cubes from exceeding base mesh bounds",
            default=True
            )
    uf = BoolProperty(
            name="Uniform Cube Quantity",
            default=False
            )
    qn = IntProperty(
            name="Cube Quantity",
            default=10,
            min=1, max=1500
            )
    mn = FloatVectorProperty(
            name="Min Scales",
            default=(0.1, 0.1, 0.1),
            subtype='XYZ'
            )
    mx = FloatVectorProperty(
            name="Max Scales",
            default=(2.0, 2.0, 2.0),
            subtype='XYZ'
            )
    lo = FloatVectorProperty(
            name="XYZ Offset",
            default=(0.0, 0.0, 0.0),
            subtype='XYZ'
            )
    rsd = FloatProperty(
            name="Random Seed",
            default=1
            )

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj is not None and obj.type == "MESH" and obj.mode == "OBJECT"

    def draw(self, context):
        layout = self.layout

        box = layout.box()
        box.label(text="Options:")
        box.prop(self, "dc")
        box.prop(self, "wh")
        box.prop(self, "uf")

        box = layout.box()
        box.label(text="Parameters:")
        box.prop(self, "qn")
        box.prop(self, "mn")
        box.prop(self, "mx")
        box.prop(self, "lo")
        box.prop(self, "rsd")

    def execute(self, context):
        rsdchange = self.rsd
        oblst = []
        uvyes = 0
        bpy.ops.group.create(name='Cubagrouper')
        bpy.ops.group.objects_remove()

        for ob in bpy.context.selected_objects:
            oblst.append(ob)

        for obj in oblst:
            bpy.ops.object.select_pattern(pattern=obj.name)  # Select base mesh
            bpy.context.scene.objects.active = obj
            if obj.data.uv_layers[:] != []:
                uvyes = 1
            else:
                uvyes = 0
            bpy.ops.object.group_link(group='Cubagrouper')
            dim = obj.dimensions
            rot = obj.rotation_euler
            if self.uf is True:
                area = dim.x * dim.y * dim.z
            else:
                area = 75

            for cube in range(round((area / 75) * self.qn)):
                random.seed(rsdchange)
                pmn = self.mn  # Proxy values
                pmx = self.mx
                if self.wh is True:
                    if dim.x < pmx.x:  # Keeping things from exceeding proper size
                        pmx.x = dim.x
                    if dim.y < pmx.y:
                        pmx.y = dim.y
                    if dim.z < pmx.z:
                        pmx.z = dim.z
                if 0.0 > pmn.x:  # Keeping things from going under zero
                    pmn.x = 0.0
                if 0.0 > pmn.y:
                    pmn.y = 0.0
                if 0.0 > pmn.z:
                    pmn.z = 0.0
                sx = (random.random() * (pmx.x - pmn.x)) + pmn.x  # Just changed self.mx and .mn to pmx.
                sy = (random.random() * (pmx.y - pmn.y)) + pmn.y
                sz = (random.random() * (pmx.z - pmn.z)) + pmn.z
                if self.wh is True:  # This keeps the cubes within the base mesh
                    ex = (random.random() * (dim.x - sx)) - ((dim.x - sx) / 2) + obj.location.x
                    wy = (random.random() * (dim.y - sy)) - ((dim.y - sy) / 2) + obj.location.y
                    ze = (random.random() * (dim.z - sz)) - ((dim.z - sz) / 2) + obj.location.z
                elif self.wh is False:
                    ex = (random.random() * dim.x) - (dim.x / 2) + obj.location.x
                    wy = (random.random() * dim.y) - (dim.y / 2) + obj.location.y
                    ze = (random.random() * dim.z) - (dim.z / 2) + obj.location.z
                bpy.ops.mesh.primitive_cube_add(
                            radius=0.5, location=(ex + self.lo.x, wy + self.lo.y, ze + self.lo.z)
                            )
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.transform.resize(
                    value=(sx, sy, sz), constraint_axis=(True, True, True),
                    constraint_orientation='GLOBAL', mirror=False, proportional='DISABLED',
                    proportional_edit_falloff='SMOOTH', proportional_size=1, release_confirm=True
                    )
                bpy.ops.object.mode_set(mode='OBJECT')
                select = bpy.context.object  # This is used to keep something selected for poll()
                bpy.ops.object.group_link(group='Cubagrouper')
                rsdchange += 3
            bpy.ops.object.select_grouped(type='GROUP')
            bpy.ops.transform.rotate(
                    value=rot[0], axis=(1, 0, 0), constraint_axis=(False, False, False),
                    constraint_orientation='GLOBAL', mirror=False, proportional='DISABLED',
                    proportional_edit_falloff='SMOOTH', proportional_size=1, release_confirm=True
                    )
            bpy.ops.transform.rotate(
                    value=rot[1], axis=(0, 1, 0), constraint_axis=(False, False, False),
                    constraint_orientation='GLOBAL', mirror=False, proportional='DISABLED',
                    proportional_edit_falloff='SMOOTH', proportional_size=1, release_confirm=True
                    )
            bpy.ops.transform.rotate(
                    value=rot[2], axis=(0, 0, 1), constraint_axis=(False, False, False),
                    constraint_orientation='GLOBAL', mirror=False, proportional='DISABLED',
                    proportional_edit_falloff='SMOOTH', proportional_size=1, release_confirm=True
                    )
            bpy.context.scene.objects.active = obj  # Again needed to avoid poll() taking me down
            bpy.ops.object.make_links_data(type='MODIFIERS')
            bpy.ops.object.make_links_data(type='MATERIAL')

            if uvyes == 1:
                bpy.ops.object.join_uvs()

            bpy.ops.group.objects_remove()
            bpy.context.scene.objects.active = select

            if self.dc is True:
                bpy.context.scene.objects.unlink(obj)

        return {'FINISHED'}


def register():
    bpy.utils.register_class(makestructure)


def unregister():
    bpy.utils.unregister_class(makestructure)


if __name__ == "__main__":
    register()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

bl_info = {
    "name": "Vertex Chamfer",
    "author": "Andrew Hale (TrumanBlending)",
    "version": (0, 1),
    "blender": (2, 63, 0),
    "location": "Spacebar Menu",
    "description": "Chamfer vertex",
    "wiki_url": "",
    "category": "Mesh"}


import bpy
import bmesh
from bpy.types import Operator
from bpy.props import (
        BoolProperty,
        FloatProperty,
        )


class VertexChamfer(Operator):
    bl_idname = "mesh.vertex_chamfer"
    bl_label = "Chamfer Vertex"
    bl_description = "Tri chamfer selected vertices"
    bl_options = {'REGISTER', 'UNDO'}

    factor = FloatProperty(
            name="Factor",
            description="Size of the Champfer",
            default=0.1,
            min=0.0,
            soft_max=1.0
            )
    relative = BoolProperty(
            name="Relative",
            description="If Relative, Champfer size is relative to the edge lenght",
            default=True
            )
    dissolve = BoolProperty(
            name="Remove",
            description="Remove/keep the original selected vertices\n"
                        "Remove creates a new triangle face between the Champfer edges,\n"
                        "similar to the Dissolve Vertices operator",
            default=True
            )
    displace = FloatProperty(
            name="Displace",
            description="Active only if Remove option is disabled\n"
                        "Displaces the original selected vertices along the normals\n"
                        "defined by the Champfer edges",
            soft_min=-5.0,
            soft_max=5.0
            )

    @classmethod
    def poll(self, context):
        return (context.active_object.type == 'MESH' and
                context.mode == 'EDIT_MESH')

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "factor", text="Distance" if self.relative else "Factor")
        sub = layout.row()
        sub.prop(self, "relative")
        sub.prop(self, "dissolve")
        if not self.dissolve:
            layout.prop(self, "displace")

    def execute(self, context):
        ob = context.active_object
        me = ob.data
        bm = bmesh.from_edit_mesh(me)

        bm.select_flush(True)

        fac = self.factor
        rel = self.relative
        dissolve = self.dissolve
        displace = self.displace

        for v in bm.verts:
            v.tag = False

        # Loop over edges to find those with both verts selected
        for e in bm.edges[:]:
            e.tag = e.select
            if not e.select:
                continue
            elen = e.calc_length()
            val = fac if rel else fac / elen
            val = min(val, 0.5)
            # Loop over the verts of the edge to split
            for v in e.verts:
                # if val == 0.5 and e.other_vert(v).tag:
                #    continue
                en, vn = bmesh.utils.edge_split(e, v, val)
                en.tag = vn.tag = True
                val = 1.0 if val == 1.0 else val / (1.0 - val)

        # Get all verts which are selected but not created previously
        verts = [v for v in bm.verts if v.select and not v.tag]

        # Loop over all verts to split their linked edges
        for v in verts:
            for e in v.link_edges[:]:
                if e.tag:
                    continue
                elen = e.calc_length()
                val = fac if rel else fac / elen
                bmesh.utils.edge_split(e, v, val)

            # Loop over all the loops of the vert
            for l in v.link_loops:
                # Split the face
                bmesh.utils.face_split(
                            l.face,
                            l.link_loop_next.vert,
                            l.link_loop_prev.vert
                            )

            # Remove the vert or displace otherwise
            if dissolve:
                bmesh.utils.vert_dissolve(v)
            else:
                v.co += displace * v.normal

        me.calc_tessface()

        return {'FINISHED'}


def register():
    bpy.utils.register_class(VertexChamfer)


def unregister():
    bpy.utils.unregister_class(VertexChamfer)


if __name__ == "__main__":
    register()
